#!/bin/bash

#####################################################
# This script will run vacuum analyze or 
# just analyze for any given table or list of tables.
# Only one iteration of the script will run at any
# given time.
# The script options are as follows:
#   -v vacuum
#   -a vacuum analyze
#   -vl run vacuum analyze and log
#   -al run analyze and log
# If only -l is passed by default the script will
# run an ANALYZE VERBOSE
# A sample call would be:
#   db_maint.sh -v operation_progress
#   db_maint.sh -a operation_progress backups
# Return codes for this scripts are as follows:
#   1 - bad parameters/options passed
#   2 - unable to access postgres
#   3 - error running psql command
#   4 - error found in log file
#   5 - script is already running
#   6 - psql command not found
#####################################################

script_name=${0##*/}

BPDIR=/usr/bp
present_dir=`pwd`
out_dir="${BPDIR}/logs.dir"
LOG="/dev/null"
maintenance_work_mem=200000
work_mem=100000
rc=0
run_option=""
run_verbiage="NOP"
min_pid_mem_limit=1000000
max_nproc_kill=5
tmp_file=/tmp/${script_name}.tmp
database_name="bpdb"
database_user="postgres"
table_list=""

#Try to find where psql is. If not found then exit
check_psql (){
/usr/bin/which psql > /dev/null 2>&1
rc=$?
if [[ $rc -eq 0 ]]; then
	psql_path="`/usr/bin/which psql`"
	psql_path=`echo ${psql_path%psql}`
	PSQL=`echo $psql_path/psql`
else
	printf "Could not find psql command in the path, exiting now.\n" >> $LOG
	exit 6
fi
}


# Function to check if postgres is up
check_db_is_up (){
        sudo -iu postgres pg_ctl status -D /usr/bp/data >> $LOG 2>>$LOG
        rc=$?
        echo $rc
}

# Verify only one instance of the script is running
check_script_running (){
if [ $(/sbin/pidof -x ${script_name}| wc -w) -gt 2 ]; then 
    printf "$script_name already running\n" >> $LOG
    exit 5
fi
}

# function to check the errorlog
check_log (){
        # Look for key words to see if something failed and exclude flase positives
        if (( `grep -w -e "FATAL:" -e "ERROR:" -e "not found" -e "not exist" $LOG|grep -wc -v "already exists\|PID\|--\|;\|RAISE"|tr -d ' \t\n\r\f'` > 0 )); then
                printf "An error or warning was found trying to run $run_verbiage ...\n" >> $LOG
        fi
}

run_vacuum_analyze (){
# add more tables using space separation

printf "BEGIN: `date +%F%H%M%S` \nRunning $run_verbiage for: ${table_list}\n" >> $LOG
for table_name in $table_list
do
	if [[ `check_db_is_up` -ne 0 ]]; then
		printf "Postgres seems to be down...\nExiting now...\n" >> $LOG
		printf "Please look at the $LOG log file for more details...\n" >> $LOG
		exit 2
	else
		${PSQL} -U $database_user $database_name << EOF >> $LOG 2>>$LOG
SET MAINTENANCE_WORK_MEM TO ${maintenance_work_mem} ;
$run_verbiage bp.${table_name} ;
EOF
		rc=$?
		if [[ $rc -ne 0 ]]; then
			printf "Error: Unable to execute $run_verbiage on ${table_name}\n" >> $LOG
			exit 3
		fi
		check_log
	fi
done
printf "Done with all tables: ${table_list} \nEND: `date +%F%H%M%S` \n" >> $LOG
}

# main script
USAGE="Usage: $script_name [-val] <table_list> \n -v Run vacuum analyze\n -a Run analyze\n -l Log to /usr/bp/logs.dir\n"

if (( $# < 1 )); then
    printf "Invalid option passed.\n $USAGE"
        exit 1
fi

if [[ $1 == *"a"*"v"* || $1 == *"v"*"a"* ]]; then
	printf "Invalid option passed.\n $USAGE"
	exit 1
fi

while getopts valp opt
do
	case "$opt" in
		v) run_option="v";run_verbiage="VACUUM ANALYZE VERBOSE";;
		a) run_option="a";run_verbiage="ANALYZE VERBOSE";;
		l) LOG="${out_dir}/db_maint.log";
		   touch $LOG;;
		*) printf "Invalid option passed.\n $USAGE";
		   exit 1;;
	esac
done
shift $((OPTIND-1))

check_script_running
check_psql

case "$run_option" in
	a) table_list="$*";run_vacuum_analyze;;
	v) table_list="$*";run_vacuum_analyze;;
	*) printf "Invalid option passed.\n $USAGE";
	   exit 1;;
esac 
