#!/bin/bash

#####################################################
# This script will run vacuum analyze 
# Only one iteration of the script will run at any
# given time.
#
# This script is schedule through 
# the /etc/cron.d/unitrends-pg file
#
# The script options are as follows:
#   -l run vacuum analyze and log
#   -lm do a dry-run of the maintenance
# A sample call would be:
#   db_maint_driver.sh -l
#   db_maint_driver.sh -lm
# Return codes for this scripts are as follows:
#   1 - bad parameters/options passed
#   2 - unable to access postgres
#   3 - error running psql command
#   4 - error found in log file
#   5 - scheduling mismatch
#   6 - psql command not found
#   7 - <currently not used>
#   8 - lock file found
#   9 - unable to find BPDIR directory or bputil executable
#####################################################
# A good query to run against the history table is:
#  select to_timestamp(last_maint_time) as last_maint_time,to_timestamp(start_time) as start_time,
#  to_timestamp(end_time) as end_time,
#  last_elapsed_time,relname,reltuples,expectedpages,bloatpct,autovac_freeze_pct,current_status
#  from bp.bpdb_maint_driver_hst where relname = 'backups_hash_values' order by 1 desc;
#####################################################
# Per UNIBP-12311
# maximum number of cocurrent maintenance processes default is 3, min value is 1
#####################################################

script_name=${0##*/}
USAGE="Usage: $script_name [-lm] \n -l Run and log to /usr/bp/logs.dir\n -lm Run in monitor mode and log\n"
lockFile=/var/lock/"${script_name}"-lock
#
# default schedule settings if missing from the master.ini
#
default_scheduled_day="3"
default_scheduled_hour="11"
default_max_run_time="30"
default_max_maint_process_count="3"

#
# minimum bloat percentage to run the maintenance
#
bloat_pct_threshold=20
#
# minimum autovac freeze xid consumed percentage to run the maintenance
#
autovac_theshold=80

#
# number of days to keep records in the maint_hist table
#
hst_days_keep=365

#
# postgres session parameters
#
maintenance_work_mem=200000
work_mem=100000
database_name="bpdb"
database_user="postgres"

#
# prefix of the tmp files and maintenance jobs
#
table_file_name_prefix="db_table_maint"

if [[ ! -x /usr/bp/bin/bputil ]]; then
	printf "Unable to find the bputil executable...\n Exiting now...\n"
	exit 9
fi
BPDIR=`/usr/bp/bin/bputil -g "Location Information" "BPDIR" "" /etc/default/bp.ini`
if [[  -z "$BPDIR" ]] || [[ ! -d "$BPDIR" ]]; then
	printf "Unable to find the BPDIR directory location...\n Exiting now...\n"
	exit 9
fi
database_dir=$BPDIR/data
present_dir=`pwd`
out_dir="${BPDIR}/logs.dir"
LOG="/dev/null"
rc=0
run_option=""
run_verbiage="NOP"
tmp_file=/tmp/${script_name}.tmp
current_epoch=`date '+%s'`
timeout_seconds=0
stop_epoch=`date +%s`
(( stop_epoch=stop_epoch+scheduled_run_time*60 ))
current_day=`date '+%u'`
current_hour=`date '+%k'`
current_time=`date '+%k%M'`
stop_epoch_date=`date -d @${stop_epoch}`
pending_maint_fl="t"
maint_process_count=0
table_file_name_list=""
child_pid=""
scheduled_day=`$BPDIR/bin/bputil -g "DataBase" DBMaintDayOfWeek "${default_scheduled_day}" $BPDIR/bpinit/master.ini`
scheduled_hour=`$BPDIR/bin/bputil -g "DataBase" DBMaintHour "${default_scheduled_hour}" $BPDIR/bpinit/master.ini`
scheduled_run_time=`$BPDIR/bin/bputil -g "DataBase" DBMaintMaxRuntime "${default_max_run_time}" $BPDIR/bpinit/master.ini`
max_maint_process_count=`$BPDIR/bin/bputil -g "DataBase" DBMaintMaxProcessCount "${default_max_maint_process_count}" $BPDIR/bpinit/master.ini`


#
# Depending on the postgres version the column name to
# cancel a thread is different
#
cent_ver=`/bin/uname -a`
cent_ver=`echo ${cent_ver#*el*}`
cent_ver=`echo ${cent_ver%*.*}`
if [ $cent_ver -ge 6 ]; then
        pid_column="pid"
	current_query_column="query"
	frozen_column="ceil((((age(cc.relfrozenxid)::double precision)/max_freeze_age) * 100 )::numeric(5,2)) as autovac_freeze_pct"
else
        pid_column="procpid"
	current_query_column="current_query"
	frozen_column="0::numeric(5,3) as autovac_freeze_pct"
fi



#
# Function to put a standard exit message
# when exiting and to remove the lockfile
#
cleanup (){
	printf "Exiting now...\n" >> $LOG
	printf "END: `date +%F%H%M%S` \n" >> $LOG
	/bin/rm -f $lockFile >/dev/null 2>&1
}

#
#Try to find where psql is. If not found then exit
#
check_psql (){
/usr/bin/which psql > /dev/null 2>&1
rc=$?
if [[ $rc -eq 0 ]]; then
	psql_path="`/usr/bin/which psql`"
	psql_path=`echo ${psql_path%psql}`
	PSQL=`echo $psql_path/psql`
else
	printf "Could not find psql command in the path, exiting now.\n" >> $LOG
	cleanup
	printf "Exit 6\n" >> $LOG
	exit 6
fi
}

#
# Function used to check that the return code from a psql
# statement was successful otherwise exit
#
check_sql_rc (){
	if [[ $rc -ne 0 ]]; then
		printf "Error: Unable to execute sql in the $funcname \n" >> $LOG
		cleanup
		printf "Exit 3\n" >> $LOG
		exit 3
	fi
}

#
# This function create a populate_driver_tbl postgres function
# Is used to refresh stats information for all bpdb tables
#
populate_driver_table (){
	funcname=$FUNCNAME
	${PSQL} -U $database_user $database_name <<-EOF >>$LOG 2>&1
	SET search_path TO bp;
	CREATE OR REPLACE FUNCTION populate_driver_tbl()
	RETURNS INT4 AS
	\$BODY\$ DECLARE
		res INT4;
		schema_name NAME;
		rel_name NAME;
		max_freeze_age double precision;
	BEGIN
		res:=0;

		RAISE NOTICE 'Populating the bpdb_maint_driver table with the most current stats data...';

		SELECT setting::double precision INTO max_freeze_age FROM pg_settings where name = 'autovacuum_freeze_max_age';

		-- Temp table to hold the current data
		CREATE TEMP TABLE tmp_driver_tbl (
		schemaname NAME NOT NULL,
		relname NAME NOT NULL,
		datawidth NUMERIC,
		reltuples BIGINT,
		relpages BIGINT,
		expectedpages BIGINT,
		bloatpct NUMERIC(5,2),
		autovac_freeze_pct NUMERIC(5,2),
		CONSTRAINT tmp_driver_pkey PRIMARY KEY (relname,schemaname)
		);

		-- Temp table to hold the last date maintenance ran for a given table
		-- this is populated from the bpdb_maint_driver_hist table
		CREATE TEMP TABLE tmp_max_maint_date (
		schemaname NAME,
		relname NAME,
		last_end_time BIGINT);

		-- Get the latest maint_date but exclude rows populated from a monitoring run
		INSERT INTO tmp_max_maint_date SELECT mh.schemaname,
		mh.relname,
		case when max(end_time) is null then extract(epoch from now())::bigint else max(end_time) end as last_end_time
		FROM bpdb_maint_driver_hst mh
		WHERE current_status != 'm'
		GROUP BY mh.schemaname, mh.relname;

		-- Populate the tmp_driver_tble with the most current stats
		INSERT INTO tmp_driver_tbl SELECT ut.schemaname,
		ut.relname,
		CASE WHEN bdw.datawidth IS NULL THEN 0 ELSE bdw.datawidth::bigint END,
		cc.reltuples::bigint AS reltuples,
		CASE WHEN cc.relpages IS NULL THEN 0 ELSE cc.relpages::bigint END,
		CASE WHEN CEIL(cc.reltuples::BIGINT * bdw.datawidth::DOUBLE PRECISION / CURRENT_SETTING('block_size'::text)::DOUBLE PRECISION)::BIGINT IS NULL THEN 0 ELSE
		ceil(cc.reltuples::bigint*bdw.datawidth::double precision / current_setting('block_size'::text)::double precision)::bigint end AS expectedpages,
			case when ((1 -(ceil(cc.reltuples::bigint * bdw.datawidth::double precision / current_setting('block_size'::text)::double precision))/
			(case when cc.relpages::double precision = 0 then 1 else  cc.relpages::double precision end))*100)::numeric(5,2) is null then 0 
			else
			((1 -(ceil(cc.reltuples::bigint * bdw.datawidth::double precision / current_setting('block_size'::text)::double precision))/
			(case when cc.relpages::double precision = 0 then 1 else  cc.relpages::double precision end))*100)::numeric(5,2) end AS bloatpct,
		${frozen_column}
		from pg_stat_user_tables ut
		join bloat_datawidth_bpdb bdw
		on (ut.relid=bdw.relid)
		join pg_class cc
		on (cc.oid = ut.relid)
		where ut.schemaname = 'bp';
		ANALYZE tmp_driver_tbl;

		-- Cycle through all the tables in the in the tmp_driver_tbl and insert them into the
		-- bpdb_maint_driver if not already there. Otherwise update their information.
		-- Delete the tables already processed from the tmp table as we go
		WHILE EXISTS(SELECT 1 FROM tmp_driver_tbl) LOOP 
			SELECT schemaname, relname INTO schema_name,rel_name FROM tmp_driver_tbl order by reltuples desc,autovac_freeze_pct desc, bloatpct desc limit 1;
			IF NOT EXISTS(SELECT 1 FROM bpdb_maint_driver where schemaname = schema_name and relname = rel_name) THEN
				INSERT INTO bpdb_maint_driver 
				(schemaname,
				relname,
				datawidth,
				reltuples,
				relpages,
				expectedpages,
				bloatpct,
				autovac_freeze_pct)
				SELECT 
				schemaname,
				relname,
				datawidth,
				reltuples,
				relpages,
				expectedpages,
				bloatpct,
				autovac_freeze_pct
				FROM tmp_driver_tbl WHERE schemaname = schema_name and relname = rel_name;

			ELSE
				UPDATE bpdb_maint_driver 
				SET datawidth = tmp.datawidth, reltuples = tmp.reltuples, relpages = tmp.relpages, expectedpages = tmp.expectedpages, bloatpct = tmp.bloatpct, autovac_freeze_pct=tmp.autovac_freeze_pct
				FROM ( SELECT datawidth, reltuples, relpages, expectedpages, bloatpct,autovac_freeze_pct FROM tmp_driver_tbl WHERE schemaname = schema_name and relname = rel_name) AS tmp
				WHERE bpdb_maint_driver.schemaname = schema_name AND bpdb_maint_driver.relname = rel_name;
			END IF;
			DELETE FROM tmp_driver_tbl WHERE schemaname = schema_name and relname = rel_name;
		END LOOP;

		-- Update the current_status in the bpdb_maint_driver for tables whose last maintenance run was not successful
		-- This way they are given priority
		UPDATE bpdb_maint_driver SET current_status = tmp.last_status, last_maint_time = tmp.last_end_time, last_elapsed_time = tmp.last_elapsed_time 
		FROM (SELECT case when mh.current_status in ('d','t','f','i') then 'c' else 'p' end as last_status, 
			last_end_time,
			mh.last_elapsed_time,tmh.schemaname,tmh.relname
			FROM bpdb_maint_driver_hst mh join tmp_max_maint_date tmh on (mh.schemaname=tmh.schemaname and mh.relname=tmh.relname and tmh.last_end_time=mh.end_time)) AS tmp
			WHERE
			bpdb_maint_driver.schemaname = tmp.schemaname AND bpdb_maint_driver.relname = tmp.relname;

		RETURN res;
	END;
	\$BODY\$
	LANGUAGE 'plpgsql' VOLATILE;
	SELECT populate_driver_tbl();
	ANALYZE VERBOSE bpdb_maint_driver;
--	DROP FUNCTION populate_driver_tbl();
	\q
	EOF
	rc=$?
	check_sql_rc
}

#
# This function takes all information from the bpdb_maint_driver table
# and after inserting it into the hst table it truncates it
#
populate_hst_table (){
	funcname=$FUNCNAME
	${PSQL} -U $database_user $database_name <<-EOF >>$LOG 2>&1
	SET search_path TO bp;
	CREATE OR REPLACE FUNCTION populate_hst_tbl()
	--RETURNS TABLE (rel_name name) AS
	RETURNS INT4 AS
	\$BODY\$ DECLARE
		res INT4;
	BEGIN
		res:=0;
		-- If we find records marked 'i' for in-progress go ahead and marked them as 
		-- 'f' for failed before copying them over to the hist table
		UPDATE bpdb_maint_driver SET current_status = 'f' where current_status = 'i';

		RAISE NOTICE 'Copying existing rows in bpdb_maint_driver to bpdb_maint_driver_hst...';
		INSERT INTO bpdb_maint_driver_hst 
		(schemaname, 
		relname,
		datawidth,
		reltuples, 
		relpages, 
		expectedpages, 
		bloatpct,
		autovac_freeze_pct,
		last_maint_time,
		last_elapsed_time,
		start_time, 
		end_time,
		current_status)
		SELECT schemaname, relname, datawidth, reltuples, relpages, expectedpages, 
		bloatpct,autovac_freeze_pct,last_maint_time,
		(end_time -  start_time) as last_elapsed_time,start_time,end_time,current_status 
		FROM bpdb_maint_driver;
		RAISE NOTICE 'Truncating the bpdb_maint_driver table...';
		TRUNCATE TABLE bpdb_maint_driver;
		RETURN res;
	END;
	\$BODY\$
	LANGUAGE 'plpgsql' VOLATILE;

	-- This procedure will maintain only days_no for each table in the bpdb_maint_driver_hst table
	-- This way we purge the bpdb_maint_driver_hst and prevent it from growing too big
	-- This can be improved later to make sure we keep at least X number of records per table but
	-- right now we just purge all records older than days_no
	CREATE OR REPLACE FUNCTION cleanup_maint_hst_tbl(days_no int)
	RETURNS INT4 AS
	\$BODY\$ DECLARE
		res INT4;
		min_date timestamp;
	BEGIN
		SELECT MIN(to_timestamp(last_maint_time)::timestamp) INTO min_date FROM bp.bpdb_maint_driver_hst;
		IF (SELECT date_part('day',now()- min_date)) > days_no THEN
			RAISE NOTICE 'Deleting all records for % from the bp.bpdb_maint_driver_hst table',min_date;
			DELETE FROM bpdb_maint_driver_hst WHERE to_date(to_timestamp(last_maint_time)::text,'YYYY MM DD') <= to_date(min_date::text,'YYYY MM DD');
		END IF;
		RETURN res;
	END;
	\$BODY\$
	LANGUAGE 'plpgsql' VOLATILE;

	SELECT cleanup_maint_hst_tbl(${hst_days_keep});
	SELECT populate_hst_tbl();
	-- Run a vacuum full analyze after purge
	VACUUM FULL VERBOSE bpdb_maint_driver;
	VACUUM ANALYZE VERBOSE bpdb_maint_driver_hst;
	\q
	EOF
	rc=$?
	check_sql_rc
}

#
# This function checks the master.ini file values for the db maintenance
# Make sure they are valid and it makes sure it really is time to run or not
#
check_schedule (){
	funcname=$FUNCNAME
	if [ "$scheduled_day" -ge 0 -a "$scheduled_day" -le 6 ]; then
		if (( $scheduled_day != $current_day )); then
			printf "Scheduled maintenance day is ${scheduled_day}. Current day is ${current_day}.\n" >> $LOG
			cleanup
			printf "Exit 5\n" >> $LOG
			exit 5
		fi
	else
		printf "Invalid value for DBMaintDayOfWeek= ${scheduled_day}. Valid values are between 0-6.\n" >> $LOG
		cleanup
		printf "Exit 5\n" >> $LOG
		exit 5
	fi
	if [ "$scheduled_hour" -ge 0 -a "$scheduled_day" -le 23 ]; then
		if (( $scheduled_hour != $current_hour )); then
			printf "Scheduled maintenance hour is ${scheduled_hour}. Current hour is ${current_hour}.\n" >> $LOG
			cleanup
			printf "Exit 5\n" >> $LOG
			exit 5
		fi
	else
		printf "Invalid value for DBMaintHour is ${scheduled_hour}. Valid values are between 0-23.\n" >> $LOG
		cleanup
		printf "Exit 5\n" >> $LOG
		exit 5
	fi
	if [ "$scheduled_run_time" -ge 10 -a "$scheduled_run_time" -le 59 ]; then
		if (( $current_epoch > $stop_epoch )); then
			printf "Scheduled stop time is $scheduled_time. Current time is ${current_time}.\n" >> $LOG
			cleanup
			printf "Exit 5\n" >> $LOG
			exit 5
		fi
	else
		printf "Invalid value for DBMaintMaxRuntime is ${scheduled_run_time}. Valid values are between 10-59.\n" >> $LOG
		cleanup
		printf "Exit 5\n" >> $LOG
		exit 5
	fi
	if [ "$max_maint_process_count" -lt 1 -o "$max_maint_process_count" -gt 5 ]; then
		printf "Current max process count is ${max_maint_process_count}.\n" >> $LOG
		printf "Invalid value for DBMaintMaxProcessCount. Valid values are between 1-5.\n" >> $LOG
		cleanup
		printf "Exit 5\n" >> $LOG
		exit 5
	fi
}

#
# This function create the get_maint_table_name postgres function
# The function is used to determine which tables qualify for maintenance
# from the bpdb_maint_driver table
#
create_maintenance_sp (){
	# current_status
	# p - pending maintenance
	# i - maintenance in progress
	# f - maintenance failed
	# t - maintenance terminated
	# s - successful maintenance
	# n - maintenance skipped, didn't meet the bloat percent threshold
	# d - maintenance deferred, run time would've exceeded maintenance window
	# c - critical, this indicates maintenance was not successful or terminated
	#     and in the current run it needs to be priorotize
	# m - populated for monitoring purposes only
	funcname=$FUNCNAME
	$PSQL -U $database_user $database_name -tq <<-EOF >> $LOG 2>&1
	SET search_path to bp;
	SET MAINTENANCE_WORK_MEM TO $maintenance_work_mem ;
	CREATE OR REPLACE FUNCTION get_maint_table_name(stop_epoch bigint,bloat_threshold int4,autovac_threshold int4)
	--RETURNS TABLE (rel_name name) AS
	RETURNS name AS
	\$BODY\$ DECLARE
		table_name NAME;
		bloat_pct NUMERIC(5,2);
		autovac_pct NUMERIC(5,2);
		last_time_elapsed BIGINT;
		bloat_maint BOOLEAN;
		autovac_xid_maint BOOLEAN;
	BEGIN
		last_time_elapsed:=0;
		bloat_maint:=FALSE;
		autovac_xid_maint:=FALSE;
		-- Lock the table since we will spawn multiple processes at the same time
		LOCK TABLE bpdb_maint_driver IN SHARE ROW EXCLUSIVE MODE; 

		-- Pick the first available table for maintenance giving priority
		-- based on the current_status
		SELECT relname,bloatpct,autovac_freeze_pct INTO table_name,bloat_pct,autovac_pct FROM bpdb_maint_driver 
		WHERE current_status IS NULL OR 
		current_status IN ('p','c')
		ORDER BY reltuples, current_status,autovac_freeze_pct,bloatpct limit 1;

		-- Determine if the table really qualifies for maintenance based on the
		-- bloat_threshold. If it does the update the current_status to in-progress
		IF bloat_pct < bloat_threshold THEN
			RAISE NOTICE 'Table % does not meet the minimum bloat percentage threshold of %',table_name,bloat_threshold;
		ELSE
			RAISE NOTICE 'Table % meets the minimum bloat percentage threshold of %',table_name,bloat_threshold;
			bloat_maint:=TRUE;
		END IF;

		-- Determine if the table really qualifies for maintenance based on the
		-- autovac frozen xid. If it does the update the current_status to in-progress
		IF (autovac_pct < autovac_threshold) THEN
			RAISE NOTICE 'Table % does not meet the minimum autovac frozen xid threshold of %',table_name,autovac_threshold;
		ELSE
			RAISE NOTICE 'Table % meets the minimum autovac frozen xid threshold of %',table_name,autovac_threshold;
			autovac_xid_maint:=TRUE;
		END IF;

		-- If maintenance is needed because of bloat of running out of tran xids go for it
		-- Alsothere is enough time to run the maintenance for the table
		-- based on the last run time.  If there isn't don't run the maintenance and
		-- update the current_status.
		IF bloat_maint OR autovac_xid_maint THEN
			UPDATE bpdb_maint_driver SET current_status = 'i', start_time = extract(epoch from now())::bigint
			WHERE relname = table_name;

			SELECT last_elapsed_time
			INTO last_time_elapsed FROM bpdb_maint_driver
			WHERE relname = table_name;

			IF (last_time_elapsed + extract(epoch from now())::bigint)  < stop_epoch THEN
				RAISE NOTICE 'Running VACUUM FULL ANALYZE VERBOSE for %',table_name;	
			ELSE
				RAISE NOTICE 'Not enough time to run maintenance for %', table_name;
				UPDATE bpdb_maint_driver SET current_status = 'd', start_time = extract(epoch from now())::bigint,
				end_time = extract(epoch from now())::bigint
				WHERE relname = table_name;
				table_name:=NULL;
			END IF;
		ELSE
			UPDATE bpdb_maint_driver SET current_status = 'n', start_time = extract(epoch from now())::bigint,
			end_time =  extract(epoch from now())::bigint
			WHERE relname = table_name;
			table_name:=NULL;	
		END IF;
		table_name:='bp.'||table_name;
		RETURN table_name;
	END;
	\$BODY\$
	LANGUAGE 'plpgsql' VOLATILE;
	\q 
	EOF
	rc=$?
	check_sql_rc
}

#
# This function makes sure no ongoing maintenance runs past
# the max_run_time. Otherwise it will kill the process and 
# update the status of the maint for the table in question in the
# bpdb_maint_driver table
#
check_maint_processes (){
	funcname=$FUNCNAME
	current_epoch=`date '+%s'`
	if (( $current_epoch > $stop_epoch )); then
	for pid_number in `ps -aef|grep -v grep|grep "${table_file_name_prefix}"|awk '{print $2}'`
	do
		temp_name=`ps -fp $pid_number|tail -1`
		temp_name=${temp_name##*${table_file_name_prefix}_}
		schema_name=${temp_name%%.*}
		table_name=${temp_name##*.}
		table_name=${table_name%_*}
		child_pid=`pgrep -P ${pid_number} |tail -1`
		while [[ $child_pid != "" ]]; do
			kill -15 $child_pid
			child_pid=`pgrep -P ${pid_number} |tail -1`
		done
		kill $pid_number >>$LOG 2>&1
		$PSQL -U $database_user $database_name -tqc"UPDATE bp.bpdb_maint_driver SET current_status = 't', end_time = extract(epoch from now())::bigint 
			WHERE relname = '${table_name}' and schemaname='${schema_name}';" >>$LOG 2>&1
		check_sql_rc
		sleep 5
	done
	fi
}

#
# Main function.
#
run_maint (){
	funcname=$FUNCNAME

	if [[ "${run_option}" != "m" ]]; then
		check_schedule
		# 
		# Even if the -l option is not used log acivity if schedule qualifies
		# strip the .sh from the script name for the log name
		#
		LOG="${out_dir}/${script_name%.*}.log"
		printf "BEGIN: `date +%F%H%M%S` \nRunning $run_verbiage \n" >> $LOG
		printf "Maintenance should stop by $stop_epoch_date \n" >> $LOG
	else
		# 
		# Even if the -l option is not used log acivity if schedule qualifies
		# strip the .sh from the script name for the log name
		#
		LOG="${out_dir}/${script_name%.*}.log"
		printf "BEGIN: `date +%F%H%M%S` \nRunning $run_verbiage \n" >> $LOG
	fi

	populate_hst_table
	populate_driver_table
	create_maintenance_sp

	#
	# Keep grabing tables that qualifiy for maintenance while we don't exceed the stop_epoch
	#
	current_epoch=`date '+%s'`
	while (( $current_epoch < $stop_epoch )); do
		#
		# Check to see if we still have talbes to process in the bp.bpdb_maint_driver table
		#
		pending_maint_fl=`$PSQL -U $database_user $database_name -tqc"select exists(select 1 from bp.bpdb_maint_driver where current_status in ('p','c') or current_status is null)"|head -1`
		pending_maint_fl=${pending_maint_fl# }
		if [[ "${pending_maint_fl}" == "t" ]]; then
			#
			# Execute the get_maint_table_name function to get the name of the next table to do maintenance on
			#
			table_name=`$PSQL -U $database_user $database_name -tq <<-EOF 2>>$LOG
					SET search_path to bp; 
					SELECT get_maint_table_name($stop_epoch,${bloat_pct_threshold},${autovac_theshold}); 
					\q 
					EOF`
			if (( $rc != 0 )); then
				printf "Errors found running the run_maintenane_sp function...\n" >> $LOG
				printf "Will stop running the maintenance now...\n" >> $LOG
				break	
			fi
			check_sql_rc
			table_name=${table_name## }
			if [[ "${table_name}" != "" ]]; then
				#
				# Create a bash script file in /tmp to run the maintenance on the selected table
				#
				table_file_name=/tmp/${table_file_name_prefix}_${table_name}_${RANDOM}
				table_file_name_list="${table_file_name} ${table_file_name_list}"
				/bin/touch $table_file_name
				printf "#!/bin/bash \n" > $table_file_name
				#
				# This is a clean_up function used executed if the timeout is met we will mark the maintenance as failed
				#
				printf "function clean_up {\n" >> $table_file_name
				printf "%s -U %s %s -tqc\"UPDATE bp.bpdb_maint_driver SET current_status = 't', end_time = extract(epoch from now())::bigint WHERE relname = '%s';\" \n" \
					"$PSQL" "$database_user" "$database_name" "${table_name#*bp.}" >> $table_file_name
				printf " exit \n" >> $table_file_name
				printf "}\n" >> $table_file_name

				#
				# trap TERM and execute the clean_up function
				#
				printf "trap clean_up SIGHUP SIGINT SIGTERM \n" >> $table_file_name

				printf "rc=0 \n" >> $table_file_name
				printf "%s -U %s %s -tqac\"VACUUM FULL ANALYZE VERBOSE %s\" >> $LOG 2>&1 \n" "$PSQL" "$database_user" "$database_name" "$table_name">> $table_file_name
				printf "rc=\$\? \n" >> $table_file_name
				printf "if (( \$rc == 0 )); then \n" >> $table_file_name
				printf "%s -U %s %s -tqc\"UPDATE bp.bpdb_maint_driver SET current_status = 's', end_time = extract(epoch from now())::bigint WHERE relname = '%s';\" \n" \
					"$PSQL" "$database_user" "$database_name" "${table_name#*bp.}" >> $table_file_name
				printf "else \n" >> $table_file_name
				printf "%s -U %s %s -tqc\"UPDATE bp.bpdb_maint_driver SET current_status = 'f', end_time = extract(epoch from now())::bigint WHERE relname = '%s';\" \n" \
					"$PSQL" "$database_user" "$database_name" "${table_name#*bp.}" >> $table_file_name
				printf "fi \n" >> $table_file_name
				printf "exit 0 \n" >> $table_file_name

				#
				# If the run_option was m (maintenance) then don't execute the bash file
				#
				if [[ "${run_option}" != "m" ]]; then
					/bin/chmod 755 $table_file_name
					current_epoch=`date '+%s'`
					#
					# Figure out how many seconds can the maintenance job actually run without going over the stop time
					#
					(( timeout_seconds=stop_epoch-current_epoch ))
					#
					# If the diff is negative then don't run, we went over
					# set the current status to deferred
					#
					if (( timeout_seconds > 0 )); then
						#
						# If centos6 or greater use the timeout command otherwise
						# use sleep and kill logic
						if [ $cent_ver -ge 6 ]; then
							nohup /usr/bin/timeout ${timeout_seconds}s $table_file_name >> $LOG 2>&1 &	
						else
							#
							# For cent5 send the maint script to background and spawn a second monitoring script
							# this is similar to how timeout actually works
							#
							nohup $table_file_name >> $LOG 2>&1 &
							table_file_name_pid=$!
							/bin/touch ${table_file_name}_mon
							printf "#!/bin/bash \n" > ${table_file_name}_mon
							printf "sleep ${timeout_seconds} \n" >> ${table_file_name}_mon
							printf "pid_num=`ps -ef|grep -w ${table_file_name}|grep -v grep|awk '{print $2}'` \n" >> ${table_file_name}_mon
							printf "kill \${pid_num} > /dev/null 2>&1 \n" >> ${table_file_name}_mon
							/bin/chmod 755 ${table_file_name}_mon
							nohup ${table_file_name}_mon >> $LOG 2>&1 &
						fi
					else
						$PSQL -U $database_user $database_name -tqc \ 
						"UPDATE bp.bpdb_maint_driver SET current_status = 'd', end_time = extract(epoch from now())::bigint WHERE relname = '${table_name#*bp.}';" >> $LOG 2>&1
					fi
					#
					# Don't run more then tne number of max_maint_process_count
					#
					maint_process_count=`ps -aef|grep -v grep|grep "${table_file_name_prefix}"|wc -l`
					while (( $maint_process_count > $max_maint_process_count ))
					do
						usleep 50000
					#	check_maint_processes
						maint_process_count=`ps -aef|grep -v grep|grep "${table_file_name_prefix}"|wc -l`
					done
				fi
			fi
		else
			break
		fi
		current_epoch=`date '+%s'`
	done
	
	#
	# Keep monitoring any maintenance jobs still running even after exiting the main loop
	#
	maint_process_count=`ps -aef|grep -v grep|grep "${table_file_name_prefix}"|wc -l`
	while (( $maint_process_count > 0 ))
	do
		check_maint_processes
		maint_process_count=`ps -aef|grep -v grep|grep "${table_file_name_prefix}"|wc -l`
	done

	#
	# If maint option then just update all rows current_status to m
	#
	if [[ "${run_option}" == "m" ]]; then
		$PSQL -U $database_user $database_name -tqc"UPDATE bp.bpdb_maint_driver SET current_status = 'm', end_time = extract(epoch from now())::bigint;" >> $LOG  2>&1
	fi

	#
	# If we still have any VACUUM process running kill it
	#
	$PSQL -U $database_user $database_name -tqc"SELECT pg_cancel_backend($pid_column) FROM pg_stat_activity WHERE upper($current_query_column) like 'VACUUM%'" >> $LOG 2>&1
	check_sql_rc
	if [ $cent_ver -ge 6 ]; then
		$PSQL -U $database_user $database_name -tqc"SELECT pg_terminate_backend($pid_column) FROM pg_stat_activity WHERE upper($current_query_column) like 'VACUUM%'" >> $LOG 2>&1
		check_sql_rc
	fi

	#
	# Remove tmp files and populate history taable
	#
	/bin/rm -f $table_file_name_list >>$LOG 2>&1
	populate_hst_table
}


#main script

#
# If the lock file is older than 24 hours then try to remove it
# Maybe a catastrophic event happend
#
if [[ -e ${lockFile} ]]; then
	if test "`find ${lockFile} -mmin +1440`"; then
		printf "Found ${lockFile}.  Will remove it now...\n"
		/bin/rm -f ${lockFile} &>/dev/null
	fi
fi

#
#  Make sure only one copy of this script can run at a time.
#  Only lock the file no more than 60 minutes
#
if ! lockfile -r0 -l 3600 "$lockFile"; then
	printf "${script_name}: Can't get the lock\n"
	exit 8
fi

# Make sure we can find the psql command or exit
check_psql
rc=0
if (( $# < 1 )); then
        run_maint
	cleanup
        exit $rc
fi

while getopts lm opt
do
        case "$opt" in
                m) run_option="m";
                   run_verbiage="MONITORING LOGGING";;
                l) LOG="${out_dir}/${script_name%.*}.log";
		   run_option="l";
		   run_verbiage="LOGGING";
                   touch $LOG;;
                *) printf "Invalid option passed.\n $USAGE";
		   cleanup;
                   exit 1;;
        esac
done

if [[ $1 == *"lm"* || $1 == *"ml"* ]]; then
        run_option="m"
	run_verbiage="MONITORING LOGGING"
fi

case "$run_option" in
        l) run_maint; cleanup; exit $rc;;
        m) run_maint; cleanup; exit $rc;;
	*) printf "Invalid option passed.\n $USAGE";
	   cleanup;
	   exit 1;;
esac
