#!/bin/bash

#####################################################
# This script will by default kill any transaction
# using more than 1GB of VM or idle for more than
# 1 day
# The script options are as follows:
#   -p  run through pid process list to kill
#   -pl log execution
# If only -l is passed by default the script will
# run and try to find threads that meet the threshold to
# cancel them
# A sample call would be:
#   db_pkill.sh
#   db_pkill.sh -l
#   db_pkill.sh -p <pid>
#   db_pkill.sh -l <pid>
#   db_pkill.sh -pl <pid>
# Return codes for this scripts are as follows:
#   1 - bad parameters/options passed
#   2 - unable to find psql binary
#   3 - no pid killed
#####################################################

script_name=${0##*/}
# Verify only one instance of the script is running
if [ $(/sbin/pidof -x ${script_name}| wc -w) -gt 2 ]; then 
    printf "$script_name already running\n"
    exit
fi
USAGE="Usage: $script_name [-pl] <pid_list>\n -p Process pid list\n -m Monitor only don't kill\n -l Log to /usr/bp/logs.dir\n"

BPDIR=/usr/bp
present_dir=`pwd`
out_dir="${BPDIR}/logs.dir"
LOG="/dev/null"
rc=0
min_pid_mem_limit=1000000 # in KB's
min_pid_idle_time=1 #in days
max_nproc_kill=5
run_option=""
run_verbiage="DEFAULT - Find top $max_nproc_kill postgres rss processes"
# use \| to add more process owners to exclude from kill
# this has to match the first column of the lsof output
# excl_powner="vcd\|postgres\|appliance\|restore_d\|retention\|tasker"
excl_powner="vcd"
tmp_file=/tmp/${script_name}_pid_list.tmp
database_name="bpdb"
database_user="postgres"
pid_count=0
pid_list=""
pid_port=""
pid_time=0
table_list=""
found_pid=0
/bin/uname -r|grep -q el6
is_cent6=$?

if [ $is_cent6 -eq 0 ]; then
	pid_column="pid"
else
	pid_column="procpid"
fi

find_parent_pid (){
	powner=postgres
	tmp_parent=/tmp/${script_name}_parent_$$.tmp
	if [ $pid_port -gt 0 ]; then
#		/usr/sbin/lsof|grep $pid_port > $tmp_parent 2>&1
		/sbin/fuser -v -n tcp $pid_port > $tmp_parent 2>&1
		if (( $? == 0 )); then
			cat $tmp_parent >> $LOG
			powner=`tail -1 $tmp_parent|awk '{print $5}'`
		fi
	fi
	/bin/rm -f $tmp_parent
}

pkill_cent5 (){
	# pg_cancel_backend is equivalent to sending SIGINT to the process.
	# since process is idle a kill -15 is needed to really cancel the connection
	#pg_terminate_backend likewise for SIGTERM
	if [[ `echo "$powner"|grep -c $excl_powner` -eq 0 ]]; then
		printf "Killing $pid with owner $powner \n" >> $LOG
		/bin/kill -15 $pid > /dev/null 2>&1
	else
		printf "The process owner for $pid is $powner which is in the exclude process owner list: $excl_powner \n" >> $LOG
	fi
}
pkill_cent6 (){
	if [[ `echo "$powner"|grep -c $excl_powner` -eq 0 ]]; then
		printf "Killing $pid with owner $powner \n" >> $LOG
		$PSQL -U $database_user $database_name -c "select pg_terminate_backend($pid) from pg_stat_activity where $pid_column = $pid and state = 'idle';" >>$LOG 2>&1
	else
		printf "The process owner for $pid is $powner which is in the exclude process owner list: $excl_powner \n" >> $LOG
	fi
}
kill_process (){
printf "BEGIN: `date +%F%H%M%S` \nRunning $run_verbiage \n" >> $LOG
if (( $pid_count > 0 )); then
	printf "Process list: $pid_list\n" >> $LOG
        for pid in $pid_list 
	do
		/bin/ps -fp $pid|grep -v "grep\|${script_name}" >> $LOG
		pid_port=`$PSQL -U $database_user $database_name -tAc "select client_port from pg_stat_activity where $pid_column = $pid "`
		if [[ $pid_port -eq "" ]]; then
			pid_port=0
		fi
		find_parent_pid
		$PSQL -U $database_user $database_name -tAc "select * from pg_stat_activity where $pid_column = $pid;" >>$LOG 2>&1
		if [[ "$run_option" != "m" ]]; then
			printf "Attempting to terminate process ${pid} from postgres now...\n" >> $LOG
			if [[ `$PSQL -U $database_user $database_name -tAc "select 1 from pg_stat_activity where $pid_column = ${pid} "` == "1" ]]; then
				if [ $is_cent6 -eq 0 ]; then
					pkill_cent6
				else
					pkill_cent5
				fi
			else
					printf "Was not able to find pid $pid ...\n" >> $LOG
			fi
		fi
        done
else
        /bin/ps aux |grep -i '^postgres.*idle\s*$' |sort -k6 -n -r |head -n$max_nproc_kill > $tmp_file
        while read line
        do
                rss=`echo $line |awk '{ print $6 }'`
                pct=`echo $line |awk '{ print $4 }'`
                pid=`echo $line |awk '{ print $2 }'`
                msg="pid=$pid pct=$pct rss=$rss"
                database_name="postgres"
		pid_time=`$PSQL -U $database_user $database_name -tAc "select trim(date_part('day',age(backend_start))::char) as run_time_d from pg_stat_activity where $pid_column = $pid"`
		# handle the case when the query returns a - not a numeric value
		case "$pid_time" in
			-) pid_time=0;;
			*) true ;;
		esac
                if [ $rss -gt $min_pid_mem_limit ] || [ $pid_time -gt $min_pid_idle_time ]; then
                        printf "Found $msg $pid exceeding $rss of $min_pid_mem_limit virtual memory usage or $pid_time of $min_pid_idle_time idle time in minutes threshold\n" >> $LOG
			/bin/ps -fp $pid|grep -v "grep\|${script_name}" >> $LOG
			pid_port=`$PSQL -U $database_user $database_name -tAc "select client_port from pg_stat_activity where $pid_column = $pid "`
			find_parent_pid
			$PSQL -U $database_user $database_name -tAc "select * from pg_stat_activity where $pid_column = $pid;" >>$LOG 2>&1
			found_pid=1
			if [[ "$run_option" != "m" ]]; then
				if [ $is_cent6 -eq 0 ]; then
					# el6 with pg9, use pg_cancel_backend, pid
					pkill_cent6
				else
					# el5 with pg8, use pg_cancel_backend, procpid
					pkill_cent5
				fi
			fi
                else
                        printf "Found $msg, doesn't exceed $min_pid_mem_limit virtual memory usage or $pid_time of $min_pid_idle_time idle time in minutes threshold \n" >> $LOG
                fi
        done <$tmp_file
        /bin/rm -f $tmp_file
	if (( $found_pid == 0 )); then
		printf "No pid found to exceed the virtual memory usage threshold\n" >> $LOG
		rc=3
	fi
fi
printf "Done with $run_verbiage \nEND: `date +%F%H%M%S` \n" >> $LOG

}

# main script

#Try to find where psql is. If not found then exit
/usr/bin/which psql > /dev/null 2>&1
rc=$?
if [[ $rc -eq 0 ]]; then
	psql_path="`/usr/bin/which psql`"
	psql_path=`echo ${psql_path%psql}`
	PSQL=`echo $psql_path/psql`
else
	printf "Could not find psql command in the path, exiting now.\n" >> $LOG
	exit 2
fi
rc=0

if (( $# < 1 )); then
	kill_process
	exit $rc
fi

while getopts lpm opt
do
	case "$opt" in
		p) run_option="p";
		   run_verbiage="KILLING PROCESS LIST";;
		m) run_option="m";
		   run_verbiage="MONITORING PROCESSES";;
		l) LOG="${out_dir}/${script_name}.log";
		   touch $LOG;;
		*) printf "Invalid option passed.\n $USAGE";
		   exit 1;;
	esac
done
if [[ $1 == *"pm"* || $1 == *"mp"* ]]; then
	run_option="pm"
	run_verbiage="MONITORING PROCESS LIST"
fi
shift $((OPTIND-1))

case "$run_option" in
	p) pid_list="$*";pid_count=$# ;kill_process;exit $rc;;
	m) kill_process; exit $rc;;
	pm) pid_list="$*";pid_count=$# ;run_option="m";kill_process;exit $rc;;
	*) kill_process; exit $rc;;
esac 
