#!/usr/bin/perl

use warnings;
use strict;

use POSIX 'strftime';
use File::stat;
use Data::Dumper;
use Getopt::Long;
use Fcntl qw(:flock);

use constant SECONDS_PER_DAY => 60*60*24;
use constant SECONDS_PER_HOUR => 60*60;
use constant SECONDS_PER_MINUTE => 60;

my $program_name = "remove_vmware_sdk_logs";
my $log_name = $program_name;
my $program_log_count = 100; # Maintain 100 log files

my $program_start_time;
my $program_end_time;

my $vmware_logs_path = "/tmp/vmware-root"; # TODO Check what happens for a different username
# Next 2 variables are needed because we want to delete only files
# whose names are of the format - vmware-<process_id_number>.log
my $vmware_log_name_prefix = "vmware-";
my $vmware_log_name_postfix = ".log";

# Command-line arguments
my $help = 0;
my $delete = 0;
my $minutes_old = 0;	# Delete logs that are older than $minutes_old minutes
my $hours_old = 0;	# Delete logs that are older than $hours_old hours
my $days_old = 0;	# Delete logs that are older than $days_old days

my $files_successfully_deleted = 0;
my $files_failed_to_delete = 0;

my $lock_file = "/var/run/$program_name.pid";

sub setup_logging {

	my $i = $program_log_count;
	while ($i >= 1) {
		my $m = $i + 1;
		rename "/usr/bp/logs.dir/$log_name.log.$i", "/usr/bp/logs.dir/$log_name.log.$m";
		$i--;
	}
	rename "/usr/bp/logs.dir/$log_name.log", "/usr/bp/logs.dir/$log_name.log.1";

	my $date = strftime ("%A %m/%d/%Y %H:%M:%S", localtime);

	open LOGFILE, '>', "/usr/bp/logs.dir/$log_name.log" or die $!;

	print LOGFILE "\n###############################################################################\n";
	print LOGFILE "#\n# $log_name log start - $date\n#\n";
	print LOGFILE "###############################################################################\n\n";

}

sub close_logging {

	my $date = strftime ("%A %m/%d/%Y %H:%M:%S", localtime);

	print LOGFILE "\n###############################################################################\n";
	print LOGFILE "#\n# $log_name log end - $date\n#\n";
	print LOGFILE "###############################################################################\n\n";

	close LOGFILE;

}

# Prints to the opened LOGFILE
sub debug_logging($) {

	my $inputstring = shift;

	my $date = strftime ("%m/%d/%Y %H:%M:%S", localtime);

	print LOGFILE "$date | " . $inputstring . "\n";

}

# Prints to STDOUT as well as the opened LOGFILE
sub print_and_debug_logging($) {

	my $inputstring = shift;

	my $date = strftime ("%m/%d/%Y %H:%M:%S", localtime);

	print LOGFILE "$date | " . $inputstring . "\n";
	print $date . " | " . $inputstring . "\n";
}

sub find_vprotect_process_ids {

	# Run this to find list of 'vprotect' Backups and Restores that are currently active
	my $cmd = "ps -ef | awk '(/vprotect/ && (/backup/ || /restore/)) && !/awk/ {print \$2}'";

	my @vprotect_pids = `$cmd`;

	return @vprotect_pids;

}

sub remove_file {

	my $file_name = shift;

	unlink $file_name;
	if (-e $file_name) {
		debug_logging("\t\t\tERROR: Unable to delete file '$file_name'.");
		$files_failed_to_delete++;
	} else {
		debug_logging("\t\tSuccessfully deleted file '$file_name'.");
		$files_successfully_deleted++;
	}

}

sub find_and_remove_vmware_log_files {

	my $file_name;
	my $file_info;
	my $pid_from_file_name;
	my $file_last_modify_epoch_time;
	my $file_last_modify_local_time;
	my $file_age_in_seconds;

	my $seconds_old = 0;
	my $string_for_logs = "";
	my $total_number_of_vmware_log_files = 0;
	my $remove_all_vmware_sdk_logs = 0;

	my @vprotect_info;

	# Safety check
	if ( ($minutes_old eq 0) and ($hours_old eq 0) and ($days_old eq 0) ) {
		debug_logging("WARNING: No time units for file age passed. " .
				"Will delete VMware logs that are older than " .
				"12 hours (default behavior)");
		$hours_old = 12;
	}

	# Convert the unit of time into seconds
	# Save the unit in a string (to be used later for logging)
	if ($days_old) {
		$seconds_old = $days_old * SECONDS_PER_DAY;
		$string_for_logs = "$days_old days";
	} elsif ($hours_old) {
		$seconds_old = $hours_old * SECONDS_PER_HOUR;
		$string_for_logs = "$hours_old hours";
	} else {
		# $minutes_old
		$seconds_old = $minutes_old * SECONDS_PER_MINUTE;
		$string_for_logs = "$minutes_old minutes";
	}

	@vprotect_info = find_vprotect_process_ids();
	if (scalar (@vprotect_info) eq 0) {
		debug_logging("No 'vprotect' Backup or Restore processes are running currently. " .
			"Removing all VMware VDDK logs.");
		$remove_all_vmware_sdk_logs = 1;
	} else {
		foreach (@vprotect_info) {
			chomp $_;
			# Create a log file name string using the PID of the 'vprotect' process.
			# These logs will NOT be deleted as they are being accessed currently.
			$_ = $vmware_log_name_prefix . $_ . $vmware_log_name_postfix;
		}
		debug_logging("VMware log names with PIDs: @vprotect_info");
	}

	if (chdir($vmware_logs_path)) {
		debug_logging("Successfully changed directory to $vmware_logs_path");
	} else {
		debug_logging("ERROR: Unable to change directory to $vmware_logs_path");
		goto CLEANUP;
	}

	if (opendir(DIR, ".")) {
		debug_logging("Successfully opened $vmware_logs_path");
	} else {
		debug_logging("ERROR: Unable to access $vmware_logs_path.");
		goto CLEANUP;
	}

	if ($remove_all_vmware_sdk_logs) {
		debug_logging("***** Removing ALL VMware VDDK logs under $vmware_logs_path *****");
		my @file_names = readdir(DIR);
		@file_names = grep { $_ =~ /^$vmware_log_name_prefix[0-9]+$vmware_log_name_postfix$/ } @file_names;
		$total_number_of_vmware_log_files = scalar(@file_names);

		debug_logging("Deleting these files: @file_names");
		unlink @file_names;

		debug_logging("Total number of VMware log files found = $total_number_of_vmware_log_files");
		goto CLEANUP;
	}

	# Do this implementation only if script execution ends up taking more time
	##################################################################
	# Before going through the file list, we can sort the files first
	# and then, only remove those files that are older than the time of
	# execution of this script.
	# This way, if there are lots of files that are NOT needed to be
	# deleted, then, we need not go through the complete list.
	##################################################################

	while ($file_name = readdir(DIR)) {
		# VMware logs' file names are of the format - 
		# 	vmware-<process_id_number>.log
		if ($file_name =~ /^$vmware_log_name_prefix[0-9]+$vmware_log_name_postfix$/) {

			debug_logging("\tLooking at file '$file_name'");

			# This is a VMware log file
			$total_number_of_vmware_log_files++;

			if ($file_name ~~ @vprotect_info) {
				debug_logging("Not deleting file '$file_name' " .
					"because it is currently being accessed.");
				next;
			}

			# 'stat' returns a list of items related to the status of a given file
			# More information can be found here - http://perldoc.perl.org/functions/stat.html
			$file_info = stat($file_name);

			$file_last_modify_epoch_time = $file_info->mtime;
			$file_last_modify_local_time = scalar localtime $file_last_modify_epoch_time;
			debug_logging("\tFile '$file_name' was last modified on $file_last_modify_local_time");

			$file_age_in_seconds = $program_start_time - $file_last_modify_epoch_time;

			if ($file_age_in_seconds > $seconds_old) {
				debug_logging("\t\tRemoving file '$file_name' " .
					"because it's older than $string_for_logs.");
				remove_file($file_name);
			} else {
				debug_logging("\t\tNot removing file '$file_name' " .
					"because it's newer than $string_for_logs.");
			}

		} else {
			debug_logging("Not looking at '$file_name' because it's not a VMware log file.");
		}
	}

	debug_logging("Total number of VMware log files found = $total_number_of_vmware_log_files");
	debug_logging("Number of files successfully deleted = $files_successfully_deleted");
	debug_logging("Number of files that couldn't be deleted = $files_failed_to_delete");
	debug_logging("Number of files that are currently accessed for logging " .
			"for 'vprotect' Backup or Restore = " .scalar(@vprotect_info));

	CLEANUP:
	closedir(DIR);

}

sub print_usage {

	print "\n";
	print "Usage: $program_name --delete --days=<Number of days old logs to delete>\n";
	print "(OR) $program_name --delete --hours=<Number of hours old logs to delete>\n";
	print "(OR) $program_name --delete --minutes=<Number of minutes old logs to delete>\n";
	print "\n";
	print "Example:\t/usr/bp/bin/$program_name --delete --days=8\n";
	print "\t\tThis will delete all the VMware logs under /tmp/vmware-root/ " .
		"that are older than 8 days.\n\n";

	close_logging();
	exit;

}

# For invalid set of arguments, 'print_usage()' is called, which prints
# the usage and exist the program
sub get_and_validate_args {

	debug_logging("Validating command-line arguments");

	my $argc = scalar(@ARGV);
	debug_logging("Total arguments passed: $argc");
	debug_logging("Arguments passed      : @ARGV");

	if ( ($argc eq 0) or ($argc > 2) ) {
		debug_logging("Invalid number of args. Printing command line usage and exiting.");
		print_usage();
	} else {
		GetOptions(
			'minutes=i' => \$minutes_old,
			'hours=i' => \$hours_old,
			'days=i' => \$days_old,
			'delete' => \$delete,
			'help' => \$help
			);

		debug_logging("Args => minutes_old=$minutes_old, hours_old=$hours_old, " .
			"days_old=$days_old, delete=$delete, help=$help");
		if ($help) {
			print_usage();
		} elsif ($delete eq 0) {
			debug_logging("No delete argument passed. Printing command line usage and exiting.");
			print_usage();
		} elsif ($delete) {
			if ( ($minutes_old eq 0) and
				($hours_old eq 0) and
				($days_old eq 0) ) {

				debug_logging("No information related to age of logs " .
					" to delete. Printing command line usage.");
				print_usage();
			} else {
				print_and_debug_logging("Valid set of arguments passed.");
			}
		} else {
			debug_logging("Invalid set of arguments passed. " .
				"Printing command line usage and exiting.");
			print_usage();
		}
	}

	if ($minutes_old) {
		print_and_debug_logging("Deleting VMware logs under $vmware_logs_path " .
				"that are older than $minutes_old minutes");
	} elsif ($hours_old) {
		print_and_debug_logging("Deleting VMware logs under $vmware_logs_path " .
				"that are older than $hours_old hours");
	} else {
		print_and_debug_logging("Deleting VMware logs under $vmware_logs_path " .
				"that are older than $days_old days");
	}

}

sub exit_since_already_running {

	print_and_debug_logging("$program_name is already running. " .
		"Try running this after some time.");
	debug_logging("Lock file $lock_file is locked.");
	exit;

}

###########################################
# Main program starts here
############################################

sub main {

	$program_start_time = time();

	# Open a log file and start logging
	setup_logging();

	# Allow only one instance of this to run
	open(LOCK, ">>", $lock_file)
		or die "Error: Can't open lock file: $!";
	flock(LOCK, LOCK_EX|LOCK_NB)
		or exit_since_already_running();

	get_and_validate_args();

	find_and_remove_vmware_log_files();

	$program_end_time = time();

	my $execution_time = $program_end_time-$program_start_time;
	print_and_debug_logging("Time taken to execute this script: $execution_time second/s");

	print_and_debug_logging("Refer to the latest log '$program_name.log' under " .
		"'/usr/bp/logs.dir/' for more information.");

	# Close the opened log file
	close_logging();

}

exit main();
