#!/usr/bin/php
<?php

// This script is used to manage remote FLR (rFLR) as part of the self-service initiative 
// for Cloud Search/Get. RFLR are initiated by the source but prevail on the target. RESTapi's
// are used to kick-off and monitor rFLR tasks. Because RESTapi's are used, the tasks run
// as apache by default. This permission level is not sufficient to create directories and files
// in root level directories (e.g. /_Stateless/backups/D2DBackups). Therefore this script runs
// with elevated permissions per linux sudo root capabilities.
//
// NOTE: this script runs as root per the entry in /etc/sudoers
//
// There are several aspects of this script:
//   - create a unique directory to hold extracted files
//   - create a zip file for all files located within a specific directory
//   - create a zip file for all files supplied (comma separated list) on the command line
//   - add up the filesize for all files supplied (comma separated list) on the command line
//   - cleanup stale rFLR data
//
// The first several options are invoked from the webservices restore.php script. The creation of
// the unique directory is used as input to starting an asynchronous restore job. This
// information is passed back to the initiator (the source), allowing it to download the files
// in a compressed form after the restore job is complete.
// The caller of this script is responsible for ensuring the directory branch for which
// to create a new leaf contains enough free space for the restore job and subsequent zip creation.
//
// The cleanup or watchdog (-w), runs per cron, scheduled to run at 22 mins after each hour.
// The cleanup sequence looks for files and directories left around from most likely a 
// broken rFLR. Normally, when the source starts an rFLR, ultimately retrieving the files in the 
// form of a zipfile, the restored files and zipfile are purged. The unique directory hangs around
// though. Tunnels are also supported allowing extracted files zipped to be downloaded directly
// from the target via the source. These have to be purged during cleanup as well.
//
// Log files: /usr/bp/logs.dir/rflr_manage.log, rflr_manage_wd.log
//
// WARNING: most consumers of this script rely on results being written to stdout. So, beware if
// you use var_dump, or echo for debugging, this will impact the way the results are interpreted.
//

#### GLOBALS  #######

$program_name   = basename($argv[0], ".php");
$bploc          = trim(shell_exec('eval `grep BPDIR /etc/default/bp.ini`; echo $BPDIR'));
if (empty($bploc) === true) { $bploc = "/usr/bp"; }
$bpbin          = "$bploc/bin";
$loc_conn_str   = "port=5432 dbname=bpdb user=postgres connect_timeout=30";
$master_ini     = "$bploc/bpinit/master.ini";
$prog_started   = time(); # when the program started
$log_count      = 25; # watchdog runs hourly, keep a days worth, zip create runs as needed.
$log            = "";
$createDir      = false;
$runWatchdog    = false;
$zipDelFiles    = false;
$errorIfDir     = false;
$zipDir         = "";
$startDir       = "";
$debug          = false;
$filename_str   = "";
$createZip      = false;
$createZipFiles = false;
$getFileSize    = false;
$zipBaseDir     = "_rflr";  # must be in zip dir path to allow files to be deleted
$sidBaseDir     = "sid_";   # temp directories identified per system id
$orphanWait     = 24 * 60 * 60;     # 1 day for detected orphan rflr jobs to linger until purged
$lastChanceWait = 24 * 60 * 60 * 3; # 3 days detected for old stuff unclear what it is

#### GLOBALS  #######

function msg_and_exit($msg, $ret_code)
{
  global $program_name;
  global $log;
  global $pid;
  global $debug;

  # Print message and exit with return code.
  $result = array();

  if ($ret_code != 0) {
    $result['error']   = $ret_code;
    $result['message'] = $msg;
    fprintf($log, "ERROR: %s\n", $msg);
  }
  else {
    fprintf($log, "%s\n", $msg);
  }
  $msg = "### $program_name ended on ".date('Y-m-d H:i:s')."\n";
  fprintf($log, $msg);
  fclose($log);
  //exit($ret_code);
  return $result;
}

function setupLogging()
{
  global $program_name;
  global $log;
  global $log_count;
  global $runWatchdog;

  # turn off php errors while we rename files that may not exist
  error_reporting(0);

  if ($runWatchdog === true) {
    $logname = $program_name."_wd";
  }
  else {
    $logname = $program_name;
  }
  $i = $log_count;
  while ($i >= 1) {
    $m = $i + 1;
    rename("/usr/bp/logs.dir/$logname.log.$i", "/usr/bp/logs.dir/$logname.log.$m");
    $i--;
  }
  rename("/usr/bp/logs.dir/$logname.log", "/usr/bp/logs.dir/$logname.log.1");

  # set basic php reporting
  error_reporting(E_ERROR | E_WARNING | E_PARSE);

  $log = fopen("/usr/bp/logs.dir/$logname.log", "w");
  if ($log == false) {
    print "Could not open log file for writing.\n";
    exit(1);
  }

  $msg = "### $program_name started on ".date('Y-m-d H:i:s');
  fprintf($log, "%s\n", $msg);
}

function UsageAndExit()
{
  global $program_name;

  echo "Usage: $program_name <dwDh> (--create_dir <start path> | \n";
  echo "                           --create_zip <full path> <d> | \n";
  echo "                           --create_zip_files --zip_dir <zip dir> --start_dir <full path> --filenames <file_list> --error_if_dir| \n";
  echo "                           --get_file_size --filenames <file_list> | \n";
  echo "\n";
  echo "create_dir, create directory, requires a starting path,\n";
  echo "   e.g. /backups/D2DBackups/_rflr/source_118/56f95430ef697/\n";
  echo "\n";
  echo "create_zip, create a zip file for all files at the full path, requires a directory path of files to zip\n";
  echo "\n";
  echo "create_zip_files, create a zip file for specific files, requires a directory path, and a list of files to zip\n";
  echo "     if --error_if_dir is provided, the request will fail if any filename provided is a directory\n";
  echo "\n";
  echo "get_file_size, add up the size for the file list\n";
  echo "\n";
  echo "d, create zip option, DO NOT delete the files that were zipped\n";
  echo "w, start a watchdog sequence to cleanup stale files and directores\n";
  echo "D, enable debugging\n";
  echo "h, print this message and exit\n";

  exit(1);
}

function doCreateDir($newDir, $echoTrue)
{
  global $log;
  $func = "doCreateDir()";
  fprintf($log, "%s, %s\n", $func, $newDir);

  $result = array();

  # validate start dir is not null
  if ($newDir == "") {
    $result['error']   = 500; 
    $result['message'] = "Start directory path cannot be null"; 
    return $result;
  }

  # make sure the start dir has a leading and trailing slash
  $newDir  = (substr($newDir, 0, 1) == '/' ? $newDir : '/'.$newDir);
  $newDir .= (substr($newDir, -1, 1) == '/' ? '' : '/');

  # directory shouldn't exist already
  if ( (!file_exists("${newDir}")) && (!is_dir("${newDir}"))) {
    fprintf($log, "Creating new directory for rflr: %s\n", $newDir);
    $oldumask = umask(0);
    $res = mkdir($newDir, 0777, true);
    fprintf($log, "mkdir res: %s\n", $res);
    $res = chown($newDir,"apache");
    fprintf($log, "chown res: %s\n", $res);
    umask($oldumask);
  } else {
    fprintf($log, "Warning: the new directory %s for rflr already exists!\n", $newDir);
  }

  $result['new_dir'] = $newDir; 
  if ($echoTrue) {
    // flatten the results
    $flat = json_encode($result);
    echo "$flat\n";
  }
  else {
    return $result;
  }
}

function doCreateZip($startDir, $deleteAfterZip)
{
  global $log;
  global $zipBaseDir;
  global $debug;

  $func = "doCreateZip()";
  fprintf($log, "%s, %s\n", $func, $startDir);

  $zipDir = "$startDir";
  if ($debug) { fprintf($log, "%s, startDir: %s\n", $func, $zipDir); }
  # make sure the zip dir has a leading and trailing slash
  $zipDir  = (substr($zipDir, 0, 1) == '/' ? $zipDir : '/'.$zipDir);
  $zipDir .= (substr($zipDir, -1,1) == '/' ? '' : '/');

  if ($debug) { fprintf($log,"%s, zipDir: %s\n", $func, $zipDir); }
  # define the zip file
  $zipFileName = "Unitrends-Restore" . rand(0, 999) . ".zip";
  $zipFileFull = $zipDir . $zipFileName;

  // Make sure the starting path contain some structured so we don't zip the whole system
  if (!isset($zipDir) || !strstr($zipDir, $zipBaseDir)) {
    fprintf(STDERR, "Starting directory is not allowed!");
    exit(1);
  }

  # find the dirs and filelist and zip them up (zip -0 favors creation time over compression)
  # Note: find . instead of -type f or -type d picks up empty dirs too.
  # Note: ! -type l is needed to weed out symbolic links, which could be pointing to the appliance itself
  $cmd = "cd $zipDir; find . ! -type l | zip -0 $zipFileFull -@ > /dev/null 2>&1";
  if ($debug) { fprintf($log,"%s, find filelist and create zip cmd: %s\n", $func, $cmd); }
  system($cmd, $res);
  if ($res === 0) {
    # update the ownership
    $cmd = "chown apache $zipFileFull; chmod 666 $zipFileFull >/dev/null 2>&1";
    if ($debug) { fprintf($log,"%s, update ownership: %s\n", $func, $cmd); }
    system($cmd, $chownRes);
    if ($deleteAfterZip) {
      if ($debug) { fprintf($log,"%s, deleteAfterZip\n", $func); }
      // Making sure the rm -rf is contained within some structured place
      if (isset($zipDir) && strstr($zipDir, $zipBaseDir)) {
        // 1st pass: delete files, but not the new zip file (use while loop to handle dirs with spaces)
        if ($debug) { fprintf($log,"%s, 1st pass, zipDir: %s\n", $func, $zipDir); }
        $cmd = "cd $zipDir; find . -type f | grep -v $zipFileName | while read filename; do rm -rf \"\$filename\" > /dev/null 2>&1; done";
        if ($debug) { fprintf($log,"%s, delete files cmd: %s\n", $func, $cmd); }
        system($cmd, $res);

        // 2nd pass: delete directories, but not the starting dir (where the zip file resides)
        // This only cleans up the info after the uuid_1. It should delete everything from uuid_1
        // Can't delete _rflr/source_118 because it could be in use by someone else, watchdog will get it!
        if ($debug) { fprintf($log,"%s, 2nd pass, zipDir: %s\n", $func, $zipDir); }
        $dirResults = getDirs($zipDir);
        foreach ($dirResults as $fulldirpath) {
          if (is_dir("${fulldirpath}")) {
            fprintf($log,"%s, remove dirname: %s\n", $func, $fulldirpath);
            if ($debug) { fprintf($log,"%s, delete dir: %s\n", $func, $fulldirpath); }
            rmdir($fulldirpath);
          }
        }
      }
    }
    // Success
    fprintf(STDOUT, "%s",  $zipFileFull);
  } else {
    fprintf(STDERR, $res);
  }
}

function doCreateZipFiles($zipDir, $startDir, $filenames, $errorIfDir)
{
  global $log;
  global $zipBaseDir;
  global $debug;

  $func = "doCreateZipFiles()";
  fprintf($log, "%s, startDir: %s\n", $func, $startDir);
  $result = array();

  if ($debug) { fprintf($log, "%s, zipDir: %s\n", $func, $zipDir); }
  # make sure the zip dir has a leading and trailing slash
  $zipDir .= (substr($zipDir, 1) == '/' ? '' : '/');
  $zipDir .= (substr($zipDir, -1) == '/' ? '' : '/');
  if ($debug) { fprintf($log,"%s, zipDir: %s\n", $func, $zipDir); }

  $dirResults = doCreateDir($zipDir, false); // Note, the zip dir may have been created by the caller
  if (!is_array($dirResults) || !isset($dirResults['new_dir'])) {
    $result['error']   = true;
    $result['message'] = 'Download zip directory creation failed: '. $dirResults;
    fprintf($log,"%s, zipDir creation failed: %s\n", $func, $result['message']);
    return $result;
  }

  // Create the zip file, adding in 1 file at a time.
  $zipName = $zipDir . "Unitrends-Restore" . rand(0, 999) . ".zip";
  $zip     = new ZipArchive;
  $status  = $zip->open($zipName, ZipArchive::CREATE);
  if ($status != true) {
    $result['error']   = true;
    $result['message'] = 'Download zip archive creation failed: ' . $zip->getStatusString();
    fprintf($log,"%s, zipFile creation failed: %s\n", $func, $result['message']);
    return $result;
  }
  if ($debug) { fprintf($log, "zipfile %s, created\n" , $zipName); }
  $addedFile = false;
  foreach ($filenames as $filename) {
    // Validate file existence, avoid soft links
    if ( (!file_exists("${filename}")) || (is_link("${filename}")) )  {
      fprintf($log, "Skipping %s as it does not exist, or is a softlink.\n", $filename);
      continue;
    }
    if (is_dir("${filename}")) {
      // recursively get the files in the directory
      $files = new RecursiveIteratorIterator(new RecursiveDirectoryIterator("${filename}"), RecursiveIteratorIterator::SELF_FIRST);
      foreach ($files as $file) {
        if (is_file("${file}") && !is_link("${file}")) {
          if ($debug) { fprintf($log, "File in the directory: %s\n", $file); }
          // The intent of $localName is to remove the mount dir path of where the backup files reside.
          $localName = str_replace($startDir,'',$file);
          // Add current file to archive
          $status = $zip->addFile($file, $localName);
          $addedFile = true;
        }
      }
    } else {
      // The intent of $localName is to remove the mount dir path of where the backup files reside.
      // If $startDir is "", then the zipped files will contain a full path which will not match what
      // the backup really has.
      // e.g. if startDir is "": /backups/rae/303/granular/spbr0000/00000000.bak
      // e.g. if startDir is "/backups/rae/303/granular/": spbr0000/00000000.bak
      $localName = str_replace($startDir, '', $filename);
      $status = $zip->addFile($filename, $localName);
      $addedFile = true;
    }
    $statusStr = $status === true ? "true" : "false";
    if ($debug) { fprintf($log, "adding file %s with name %s to zip, status = %s\n", $filename, $localName, $statusStr); }
  }

  // Make sure at least one item was added. Empty zip files cause curl errors!
  if ($addedFile === false) {
      fprintf($log, "Request completed with no files qualified to be added.");
      // close the zip, and remove the dir
      $zip->close();
      unlink($zipName);
      $result['error']   = true;
      $result['message'] = "No files qualified (softlinks are excluded) for the download.";
      // flatten the results
      $flat = json_encode($result);
      echo "$flat\n";
      return;
  }

  // Success, close and return.
  $closeRet = $zip->close();
  if ($closeRet === false) {
    fprintf($log, "closing zipfile %s failed!", $zipName);
    $result['error']   = true;
    $result['message'] = 'Error creating zip archive: ' . $zip->getStatusString() . ".";
    // flatten the results
    $flat = json_encode($result);
    echo "$flat\n";
    return;
  }

  // success, return zip name
  $result['zipname'] = $zipName;
  fprintf($log, "zip file creation, populating was successful!\n");
  // flatten the results
  $flat = json_encode($result);
  echo "$flat\n";
}

function doGetFileSize($filenames)
{
  global $log;
  global $zipBaseDir;
  global $debug;

  $func = "doGetFileSize()";
  fprintf($log, "%s, start\n", $func);
  $result = array();

  $downloadSizeBytes = 0;
  foreach ($filenames as $filename) {
    // Valid files, non-soft links!
    if (file_exists("${filename}") && (!is_dir("${filename}")) && (!is_link("${filename}")))  {
      $downloadSizeBytes += filesize($filename);
    } else if (file_exists("${filename}") && (is_dir("${filename}")) && (!is_link("${filename}"))) {
      // recursively traverse the directory to get the size
      $files = new RecursiveIteratorIterator(new RecursiveDirectoryIterator("${filename}"), RecursiveIteratorIterator::SELF_FIRST);
      foreach ($files as $file) {
        if (is_file("${file}") && !is_link("${file}")) {
          $downloadSizeBytes += filesize($file);
          if ($debug) { fprintf($log, "Directory: %s; Size: %s\n", $file, filesize($file));}
        }
      }
    }
  }

  echo "$downloadSizeBytes";
}

function getDirs($dir, $prefix = '')
{
  $func = "getDirs()";
  $dir = rtrim($dir, '\\/');
  $result = array();

  if (file_exists("${dir}")) {
    foreach (scandir($dir) as $f) {
      if ($f !== '.' and $f !== '..') {
        if (is_dir("$dir/$f")) {
          $result[] = "$dir/$f";
          $result = array_merge($result, getDirs("$dir/$f", "$prefix$f/"));
        } else {
          //$result[] = $prefix.$f;
        }
      }
    }
  }

  arsort($result,SORT_STRING);
  return $result;
}

// Method added for tunnels to deal with dead symlinks
function getDirsAndLinks($dir, $prefix = '')
{
  // this one returns deadlinks too!
  $func = "getDirsAndLinks()";
  $dir = rtrim($dir, '\\/');
  $result = array();

  if (file_exists("${dir}")) {
    foreach (scandir($dir) as $f) {
      if ($f !== '.' and $f !== '..') {
        if (is_link("$dir/$f")) {
          $result[] = "$dir/$f";
          continue;
        }
           
        if (is_dir("$dir/$f")) {
          $result[] = "$dir/$f";
          $result = array_merge($result, getDirs("$dir/$f", "$prefix$f/"));
        } else {
          //$result[] = $prefix.$f;
        }
      }
    }
  }

  arsort($result,SORT_STRING);
  return $result;
}

function isEmptyDir($dir)
{
  $func = "isEmptyDir()";

  if (!is_dir("${dir}")) return true;
  //return (($files = @scandir($dir)) && count($files) <= 2);
  $count = `find '$dir' -type f | wc -l`;
  $count = trim($count);
  if ($count == 0) {
    return true;
  }
  return false;
}

function removeRFLRDirAndJob($dir, $localDbConn)
{
  global $log;
  global $debug;
  global $zipBaseDir;

  $func = "removeRFLRDirAndJob()";
  fprintf($log, "%s, start.\n", $func);

  $query = "DELETE FROM bp.rflr_jobs WHERE directory = '$dir'";
  fprintf($log, "%s, delete from query: %s\n", $func, $query);
  $retval = pg_query($localDbConn, $query);
  if (!$retval or pg_affected_rows($retval) == 0) {
    // entry not found, log an error
    fprintf($log, "%s, Not able to locate the rflr job per the directory %s\n", $func, $dir);
  }

  // make sure the starting path contain some structured so we don't delete the whole system
  if (!strstr($dir, $zipBaseDir)) {
    fprintf($log, "%s, rflr directory %s is invalid format. It will not be purged!\n", $func, $dir);
    return;
  }
  // Purge the directory (and all of its content)
  if ($debug) { fprintf($log,"%s, rm -rf dir: %s\n", $func, $dir); }
  `rm -rf '$dir'`;
}

function isDirInRflrJobs($dir, $localDbConn)
{
  global $log;
  global $debug;
  global $zipBaseDir;

  $func = "isDirInRflrJobs()";
  fprintf($log, "%s, start.\n", $func);

  $query = "SELECT status FROM bp.rflr_jobs WHERE directory = '$dir'";
  fprintf($log, "%s, query: %s\n", $func, $query);
  $retval = pg_query($localDbConn, $query);
  if (!$retval or pg_num_rows($retval) == 0) {
    return false;
  }

  // found it
  return true;
}

function isRestoreJobActive($localDbConn, $job_no)
{
  global $log;
  global $debug;
  global $zipBaseDir;

  $func = "isRestoreJobActive()";
  fprintf($log, "%s, start.\n", $func);

  $query = "SELECT percent_done FROM bp.jobs WHERE job_no = $job_no";
  fprintf($log, "%s, query: %s\n", $func, $query);
  $retval = pg_query($localDbConn, $query);
  // if the job exists, in any form, indicate as such.
  // Note: don't base the results on percent complete as the side effects will mess
  // up reporting the jobs/rflr_jobs relationship.
  if (!$retval or pg_num_rows($retval) == 0) {
    return false;
  }
  $row = pg_fetch_row($retval);
  $percent = $row[0];
  fprintf($log, "%s, job %s, percent: %s\n", $func, $job_no, $percent);

  return true;
}

function handleStartState($localDbConn)
{
  global $log;
  global $debug;
  global $zipBaseDir;
  global $orphanWait;

  // rflr_jobs status is 'start'
  // UC3: if restore job, continue
  // UC4: no restore job, if the dir is empty, cleanup
  // UC5: no restore job, the dir is not empty, how long to wait for a download op?

  $func = "handleStartState()";
  fprintf($log, "%s, start.\n", $func);

  // get list of jobs from rflr_jobs with status of 'start'
  $status = 'start';
  $query = "SELECT directory, job_no FROM bp.rflr_jobs WHERE status = '$status'";
  fprintf($log, "%s, query: %s\n", $func, $query);
  $retval = pg_query($localDbConn, $query);
  if (!$retval) {
    fprintf($log, "%s, query: failed, %s\n", $func, $retval);
    return;
  }
  if (pg_num_rows($retval) == 0) {
    fprintf($log, "%s, no starting jobs\n", $func);
    return;
  }
  $resultArr = pg_fetch_all($retval);
  foreach($resultArr as $result) {
    if ($debug) { $x=json_encode($result);fprintf($log, "%s, results: %s\n", $func, $x); }
    $dir    = $result['directory'];
    $job_no = $result['job_no'];

    if (isRestoreJobActive($localDbConn, $job_no) === true) {
      fprintf($log, "%s, active restore job %s, continue...\n", $func, $job_no);
      continue;
    }

    if (isEmptyDir($dir) === true) {
      // found a stale job, empty dir, clean up 
      fprintf($log, "%s, removing orphaned start job %s that had an empty dir %s\n", $func, $job_no, $dir);
      removeRFLRDirAndJob($dir, $localDbConn);
    }
    else {
      // not empty, no restore job, where is the download? 
      // wait allotted time before cleaning up
      $timeNow = time();
      $dirTime = filectime($dir);
      $msg = date('Y-m-d H:i:s', $dirTime);
      if ($debug) { fprintf($log,"%s, dir: %s, dir_time: %s\n", $func, $dir, $msg); }
      if ($timeNow - $dirTime > $orphanWait) {
        // time to clean up
        fprintf($log, "%s, removing orphaned start job %s and dir %s, expired %s\n", $func, $job_no, $dir, $msg);
        removeRFLRDirAndJob($dir, $localDbConn);
      } 
      else {
        // not yet!!
        $msg = date('Y-m-d H:i:s', ($timeNow + $orphanWait));
        fprintf($log, "%s, found a potential orphaned start job %s and dir %s, expires at %s\n", $func, $job_no, $dir, $msg);
      } 
    }
  }
}

function handleDownloadingState($localDbConn)
{
  global $log;
  global $debug;
  global $zipBaseDir;

  // rflr_jobs status is 'downloading'
  // UC0: corresponding jobs entry
  // UC1: and the dir is empty
  // UC2: and the dir is not empty

  $func = "handleDownloadingState()";
  fprintf($log, "%s, start.\n", $func);

  // get list of jobs from rflr_jobs with status of 'downloading'
  $status = 'downloading';
  $query = "SELECT directory,rflr_jobs.job_no as rjob_no,jobs.job_no as jjob_no FROM bp.rflr_jobs LEFT JOIN bp.jobs USING (job_no) WHERE rflr_jobs.status = '$status'";
  fprintf($log, "%s, query: %s\n", $func, $query);
  $retval = pg_query($localDbConn, $query);
  if (!$retval) {
    fprintf($log, "%s, query: failed, %s\n", $func, $retval);
    return;
  }
  if (pg_num_rows($retval) == 0) {
    fprintf($log, "%s, no downloading jobs\n", $func);
    return;
  }

  $resultArr = pg_fetch_all($retval);
  foreach($resultArr as $result) {
    if ($debug) { $x=json_encode($result);fprintf($log, "%s, results: %s\n", $func, $x); }
    $dir     = $result['directory'];
    $rjob_no = $result['rjob_no'];
    $jjob_no = $result['jjob_no'];
    if ($jjob_no != "" ) {
      // while the real job lingers, leave this one as well in order to allow a jobs per source relationships
      fprintf($log, "%s, found a corresponding job %s in the jobs table, letting rflr job linger until that one is complete.\n", $func, $jjob_no);
      continue;
    }
    if (isEmptyDir($dir) === true) {
      // found a completed job, clean up 
      fprintf($log, "%s, found a completed downloading job %s empty dir %s. Removing it.\n", $func, $rjob_no, $dir);
      removeRFLRDirAndJob($dir, $localDbConn);
    }
    else {
      // not empty, orphan or active? Let the last chance scheme clean it up
      fprintf($log, "%s, found a job %s which may be actively downloading or orphaned, %s. Leaving it.\n", $func, $rjob_no, $dir);
    }
  }
}
  
function cleanTempDirs($localDbConn)
{
  global $log;
  global $debug;
  global $zipBaseDir;

  // UC6: lingering tmp directories, no rflr_jobs,
  // if empty, candidates for cleanup

  $func = "cleanTempDirs()";
  fprintf($log, "%s, start.\n", $func);

  // find all the devices, then any _rflr directories
  $query = "SELECT dev_rw_name FROM bp.devices WHERE is_online = 't'";
  fprintf($log, "%s, query: %s\n", $func, $query);
  $retval = pg_query($localDbConn, $query);
  if (!$retval) {
    fprintf($log, "%s, query: failed, %s\n", $func, $retval);
    return;
  }
  if (pg_num_rows($retval) == 0) {
    fprintf($log, "%s, no devices found\n", $func);
    return;
  }

  $deviceArray = pg_fetch_all($retval);
  foreach($deviceArray as $device) {
    if ($debug) { $x=json_encode($device);fprintf($log, "%s, device: %s\n", $func, $x); }
    $device_name     = $device['dev_rw_name'];
    $device_temp_dir = $device_name . $zipBaseDir;
    $dirResults      = getDirs($device_temp_dir);

    foreach ($dirResults as $dir) {
      if ($debug) { fprintf($log,"%s, tmp dir: %s\n", $func, $dir); }
      // empty directory?
      if (isEmptyDir($dir) === true) {
        // does NOT exist in rflr table?
        if (isDirInRflrJobs($dir,$localDbConn) === false) {
          // remove the directory
          fprintf($log, "%s, directory %s was empty, non-existent in rflr_jobs table, therefore purged.\n", $func, $dir);
          `rm -rf '$dir'`;
        }
        else {
          // empty dir, exist in rflr table, leave it, UC 1 or 4 will handle.
        }
      }
      else {
        // directory is not empty
        //    - and does not exist in rflr table, it could be a spin up download -- when to remove, Last Chance to handle?
        //    - and does exist in rflr table, leave it, could be active.
      }
    }
  }
}

function purgeLastChance($localDbConn)
{
  global $log;
  global $debug;
  global $zipBaseDir;
  global $sidBaseDir;
  global $lastChanceWait;

  // remove any really old lingering tmp directories and any rflr jobs
  // This also cleans up an manual-download dirs and rflr jobs due to expiration

  $func = "purgeLastChance()";
  fprintf($log, "%s, start.\n", $func);

  // find all the devices, then any _rflr directories
  $query = "SELECT dev_rw_name FROM bp.devices WHERE is_online = 't'";
  fprintf($log, "%s, query: %s\n", $func, $query);
  $retval = pg_query($localDbConn, $query);
  if (!$retval) {
    fprintf($log, "%s, query: failed, %s\n", $func, $retval);
    return;
  }
  if (pg_num_rows($retval) == 0) {
    fprintf($log, "%s, no devices found\n", $func);
    return;
  }

  $deviceArray = pg_fetch_all($retval);
  foreach($deviceArray as $device) {
    if ($debug) { $x=json_encode($device);fprintf($log, "%s, device: %s\n", $func, $x); }
    $device_name     = $device['dev_rw_name'];
    $device_temp_dir = $device_name . $zipBaseDir;
    $dirResults      = getDirs($device_temp_dir);

    // wait allotted time before cleaning up
    $timeNow = time();

    foreach ($dirResults as $dir) {
      $dirTime = filectime($dir);
      $msg = date('Y-m-d H:i:s', $dirTime);
      if ($debug) { fprintf($log,"%s, dir: %s, dir_time: %s\n", $func, $dir, $msg); }
      if ($timeNow - $dirTime > $lastChanceWait) {
        // time to clean up
        // Special cases to deal with...if this directory is a _rflr/sid_x, then only delete it if it is empty
        $baseDirName = basename($dir);
        if ( (strstr($baseDirName, $sidBaseDir)) && (isEmptyDir($dir) === false) ) {
          fprintf($log, "%s, Not removing expired dir %s, it is a root and not empty.\n", $func, $dir);
          continue;
        }
        fprintf($log, "%s, removing orphaned start job and dir %s, expired %s\n", $func, $dir, $msg);
        removeRFLRDirAndJob($dir, $localDbConn);
      } 
      else {
        // not yet!!
        $msg = date('Y-m-d H:i:s', ($dirTime + $lastChanceWait));
        fprintf($log, "%s, found a potential orphaned dir %s, expires at %s\n", $func, $dir, $msg);
      } 
    }
  }
}

//********************************************************//
//***  Methods to handle tunnels  ************************//
//********************************************************//

function isDlTunnelRunning($source_id)
{
  global $log;
  global $debug;

  $func = "isDlTunnelRunning()";

  $source_id = $source_id + 0; // cast to int
  fprintf($log, "%s, start, system_id: %d\n", $func, $source_id);
  $res = bp_get_proxy_tunnel_path($source_id);
  if ($res === false) {
    return false;
  }

  return true;
}

function isDlUrlRunning($source_id)
{
  global $log;
  global $debug;

  $func = "isDlUrlRunning()";
  fprintf($log, "%s, start, system_id: %d\n", $func, $source_id);
  $res = bp_get_dl_url($source_id);
  if ($res === false) {
    return false;
  }

  return true;
}

function get_ini_bool_value($section, $field)
{
  global $log;
  $func = "get_ini_bool_value()";

  $v = bp_get_ini_value($section, $field);
  fprintf($log, "%s, get_ini_value res: %s::%s %s\n", $func, $section, $field, $v);
  if (strcasecmp($v, "True") == 0) { return true; }
  if (strcasecmp($v, "Yes") == 0) { return true; }
  if ($v == 1) { return true; }

  return false;
}

function findDeviceTempDir($localDbConn, $sid)
{
  global $log;
  $func = "findDeviceTempDir()";

  // NOTE:::this code has to stay in sync with includes/restore.php!!!

  // This method returns:
  // - valid results should include 'sid_rflr_dir'
  // - invalid results should include 'message'

  $result = array();

  // Find the Device Path per this src id
  $filter = array();
  $filter["system_ids"] = (int)$sid;
  $system_data_md = rest_get_systems(true,$filter);
  // Find our result, should only be one
  foreach ($system_data_md as $sub) {
    if (is_array($sub)) {
      if ($sub['id'] == $sid) {
        $system_data = $sub;
        break;
      }
    }
  }
  if (!is_array($system_data)) {
    // no source id found, log and return an error
    $errorMessage = 'Unable to locate an appliance with id '.$sid;
    $result['message'] = $errorMessage;
    return $result;
  }
  $device_id = -1;
  if (is_array($system_data)) {
    // is device set?
    if (isset($system_data['device_id'])) {
      $device_id = (int)$system_data['device_id'];
    }
  }
  fprintf($log, "%s, device_id %d\n", $func, $device_id);

  if ($device_id === -1) {
    // get the default device no.
    $def_sid = 1;
    $device_id = (int) bp_get_ini_value("Configuration Options", "DefaultTapeDevice", $def_sid);
    if ($device_id === false) {
      $device_id = -1;
    }
  }
  fprintf($log, "%s, device_id now: %d\n", $func, $device_id);

  if ($device_id === -1) {
    // no device, not much we can do, log and return an error
    $errorMessage = "Not able to determine disk device per the system id of:".$sid;
    $result['message'] = $errorMessage;
    return $result;
  }

  // Get Device info (name, online, and free space)

  // Find the device state, free space, and name.
  $query = "SELECT s.is_online,s.mb_free,d.dev_rw_name FROM bp.storage s LEFT JOIN bp.devices d ON (s.storage_id=d.storage_id) WHERE d.device_no=$device_id";
  $retval = pg_query($localDbConn, $query);
  if (!$retval or pg_num_rows($retval) == 0) {
    // no device, not much we can do, log and return an error
    $errorMessage = "Not able to determine disk device info per the device id of:".$device_id;
    $result['message'] = $errorMessage;
    return $result;
  }

  // Get the results and check values
  $row = pg_fetch_row($retval);
  $online         = $row[0];
  $mb_free        = $row[1];
  $device_name    = $row[2];
  $device_name_wc = $device_name."%"; // dir entry in db starts with the device name

  // Check values
  // online?
  if ($online != true) {
    // device is off-line, log and return an error
    $errorMessage = "The disk device ".$device_name." is offline.";
    $result['message'] = $errorMessage;
    return $result;
  }

  // Device checks passed, construct an rflr sid path
  // e.g. /backups/D2DBackups/_rflr/sid_118/
  $sid_rflr_dir = $device_name . '_rflr/sid_' . $sid . '/';
  fprintf($log, "%s, download rflr sid dir: %s\n", $func, $sid_rflr_dir);

  // valid results include 'device_temp_dir'
  $result['sid_rflr_dir'] = $sid_rflr_dir; // needed for tunnels
  return $result;
}

function createTgtDlTunnel($localDbConn, $source_id)
{
  global $log;
  global $debug;

  $func = "createTgtDlTunnel()";

  $source_id = $source_id + 0; // cast to int
  fprintf($log, "%s, start, system_id: %d\n", $func, $source_id);
  // need the directory in order to find the target side mount point
  $ftdResults = findDeviceTempDir($localDbConn, $source_id);
  if (!isset($ftdResults['sid_rflr_dir'])) {
    $errorMessage = $ftdResults['message'];
    fprintf($log, "%s, target-side tunnel creation failed for system_id %d, %s.\n", $func, $source_id, $errorMessage);
    return false;
  }
  $sid_rflr_dir = $ftdResults['sid_rflr_dir'];

  $source_side = false;
  $res = bp_create_proxy_tunnel($source_id, $sid_rflr_dir, $source_side);
  if ($res === false) {
    fprintf($log, "%s, target-side tunnel creation failed for system_id %d\n", $func, $source_id);
    return false;
  }

  fprintf($log, "%s, target-side tunnel created for system_id %d.\n", $func, $source_id);
  return true;
}

function handleTargetDlTunnels($localDbConn)
{
  global $log;
  global $debug;

  // UC0: make sure tunnels are up and running

  $func = "handleTargetDlTunnels()";
  fprintf($log, "%s, start.\n", $func);

  // for each replicated source, validate an active tunnel exists

  $query = "SELECT system_id FROM bp.systems WHERE role LIKE '%Replication Source'";
  fprintf($log, "%s, query: %s\n", $func, $query);
  $retval = pg_query($localDbConn, $query);
  if (!$retval) {
    fprintf($log, "%s, query: failed, %s\n", $func, $retval);
    return;
  }
  if (pg_num_rows($retval) == 0) {
    fprintf($log, "%s, no replication sources found.\n", $func);
    return;
  }

  $resultArr = pg_fetch_all($retval);
  foreach($resultArr as $result) {
    if ($debug) { $x=json_encode($result);fprintf($log, "%s, results: %s\n", $func, $x); }
    $system_id = $result['system_id'];
    $running = isDlTunnelRunning($system_id);
    if ($running === false) {
       $created = createTgtDlTunnel($localDbConn, $system_id);
       if ($created === true) {
         $running = isDlTunnelRunning($system_id);
       }
    }

    if ($running === true) {
      fprintf($log, "%s, target-side tunnel for system_id %d IS running.\n", $func, $system_id);
    }
    else {
      fprintf($log, "%s, target-side tunnel for system_id %d IS NOT running.\n", $func, $system_id);
    }
  }
}

function destroyAllTunnels($localDbConn, $type)
{
  global $log;
  global $debug;

  // UC0: remove any active tunnels on the source
  // UC1: remove any active tunnels on the target

  $func = "destroyAllTunnels()";
  fprintf($log, "%s, start, %s.\n", $func, $type);

  if ($type == 'source') {
    // checking source tunnels...
    // for each localhost, teardown an active tunnel
    $query = "SELECT system_id FROM bp.systems WHERE is_local_system ='t'";
    fprintf($log, "%s, query: %s\n", $func, $query);
    $retval = pg_query($localDbConn, $query);
    if ($retval  && pg_num_rows($retval) > 0) {
      $resultArr = pg_fetch_all($retval);
      foreach($resultArr as $result) {
        if ($debug) { $x=json_encode($result);fprintf($log, "%s, results: %s\n", $func, $x); }
        $system_id = $result['system_id'];
        $running = isDlTunnelRunning($system_id);
        if ($running === true) {
          $destroyed = bp_destroy_proxy_tunnel($system_id);
          if ($destroyed === true) {
            fprintf($log, "%s, source-side tunnel for system_id %d removed.\n", $func, $system_id);
          }
          else {
            fprintf($log, "%s, source-side tunnel for system_id %d NOT removed.\n", $func, $system_id);
          }
        }
      }
    }
    else {
      fprintf($log, "%s, query: failed, or no source results, %s\n", $func, $retval);
    }
  }

  if ($type == 'target') {
    // checking target tunnels...
    // for each replicated source, teardown an active tunnel
    $query = "SELECT system_id FROM bp.systems WHERE role LIKE '%Replication Source'";
    fprintf($log, "%s, query: %s\n", $func, $query);
    $retval = pg_query($localDbConn, $query);
    if ($retval  && pg_num_rows($retval) > 0) {
      $resultArr = pg_fetch_all($retval);
      foreach($resultArr as $result) {
        if ($debug) { $x=json_encode($result);fprintf($log, "%s, results: %s\n", $func, $x); }
        $system_id = $result['system_id'];
        $running = isDlTunnelRunning($system_id);
        if ($running === true) {
          $destroyed = bp_destroy_proxy_tunnel($system_id);
          if ($destroyed === true) {
            fprintf($log, "%s, target-side tunnel for system_id %d removed.\n", $func, $system_id);
          }
          else {
            fprintf($log, "%s, target-side tunnel for system_id %d NOT removed.\n", $func, $system_id);
          }
        }
      }
    }
    else {
      fprintf($log, "%s, query: failed, or no target results, %s\n", $func, $retval);
    }
  }
}

function removeOrphanedTunnel($sid)
{
  global $log;
  global $debug;

  $func = "removeOrphanedTunnel()";
  fprintf($log, "%s, start.\n", $func);

  $sid = $sid + 0; // cast to int

  // Assumption is the caller determined the tunnel was an orphaned and not running.
  // Clean things up by creating it temporarily in order to get an /var/run state
  // file, for which it can be removed.
  fprintf($log, "%s, forcing destroying of target-side tunnel for system_id %d.\n", $func, $sid);
  $sid_rflr_dir= "/tmp/_rflr/sid_".$sid."/";
  $oldumask = umask(0);
  mkdir($sid_rflr_dir, 0777, true);
  umask($oldumask);
  $source_side = false;
  bp_create_proxy_tunnel($sid, $sid_rflr_dir, $source_side);
  bp_destroy_proxy_tunnel($sid);
  rmdir($sid_rflr_dir);
  rmdir("/tmp/_rflr/");

  return;
}

function destroyOrphanedTunnels($localDbConn, $force=false)
{
  global $log;
  global $debug;

  // Orphan Tunnels can exist on a target when a replicating source is removed. An entry will exist in
  // smb.conf. If caught before a reboot, then the proxy tunnel is still up, otherwise it is no longer visible.
  // Orphan tunnels can also exist on a target when create download tunnels is changed to false, and a reboot
  // occurs. This ends up with a non-running tunnel with an entry in smb.conf for valid sids.

  // UC0: remove tunnels found in smb.conf file that do not have a valid system id ($force=false)
  // UC1: remove valid system id tunnels found in smb.conf file if not running ($force=true)

  $func = "destroyOrphanedTunnels()";
  fprintf($log, "%s, start, force=%d.\n", $func, $force);

  $resultValidSids = array();
  $query = "SELECT system_id FROM bp.systems WHERE role LIKE '%Replication Source'";
  $retval = pg_query($localDbConn, $query);
  if ($retval  && pg_num_rows($retval) > 0) {
     $resultArr = pg_fetch_all($retval);
  }
  $resultValidSids = array();
  // just the sids
  foreach($resultArr as $entry) {
    if ($debug) { $x=json_encode($entry);fprintf($log, "%s, results: %s\n", $func, $x); }
    $resultValidSids[] = $entry['system_id'];
  }

  // grep smb.conf looking for "rflr_tunnel_dl_sid_"
  $smbconf = "/etc/samba/smb.conf";
  $needle  = "rflr_tunnel_dl_sid_"; // matches cmc_proxy_mount
  $smbRes  = trim(`grep $needle $smbconf`);
  $smbArr = array();
  // If no entries, skip this, otherwise you get an array[0] = "";
  if ($smbRes != ""){
    $smbArr  = explode("\n",$smbRes);
  }
  foreach ($smbArr as $smbEntry)
  {
    // only care about the trailing sid number
    $index  = strpos($smbEntry, $needle) + strlen($needle);
    $sid = substr($smbEntry, $index);
    // get rid of any trailing ']'
    $sid = trim($sid,']');

    fprintf($log, "%s, processing sid %d.\n", $func, $sid);
    // never remove a valid, running tunnel
    // if sid is not valid remove it regardless of force setting
    if (in_array($sid,$resultValidSids) === false){
      fprintf($log, "%s, removing orphaned sid %d.\n", $func, $sid);
      if (isDlTunnelRunning($sid) === true) {
        fprintf($log, "%s, orphaned sid %d was running.\n", $func, $sid);
        bp_destroy_proxy_tunnel($sid);
      }
      else {
        fprintf($log, "%s, orphaned sid %d was not running.\n", $func, $sid);
        removeOrphanedTunnel($sid);
      }
    }
    // sid is valid, only remove it is force is true and it is not running
    elseif ( ($force === true) && (isDlTunnelRunning($sid) === false) ) {
      fprintf($log, "%s, force set, orphaned sid %d was not running.\n", $func, $sid);
      removeOrphanedTunnel($sid);
    }
  }
}

function checkTargetDlTunnelAdapter($localDbConn)
{
  global $bpbin;
  global $log;
  global $debug;

  $func = "checkTargetDlTunnelAdapter()";
  fprintf($log, "%s, start.\n", $func);
  // UC0: if openvpn is configured, validate running tunnel contains the private address
  // ToDo UC1: validate the tunnel contains the correct IP per ethernet adapters, if an IP was changed

  $openvpn = `$bpbin/cmc_openvpn server confirm`;
  if ($openvpn === false) {
    fprintf($log, "%s, openvpn not configured.\n", $func);
    return;
  }
  fprintf($log, "%s, openvpn is configured.\n", $func);

  // get the tunnel's ip, assume one tunnel
  $adp = `$bpbin/cmc_network adapter list | grep tun`;
  if ($adp === false || $adp == '') {
    fprintf($log, "%s, no tunnel adapter found.\n", $func);
    return;
  }
  $adp = preg_replace('~[\r\n]+~', '', $adp);
  fprintf($log, "%s, tunnel adapter found, $adp.\n", $func);
  $tunnel_ip = `ifconfig $adp | awk -F':' '/inet addr/{print \$2}'| sed -e 's/  P-t-P//g' |tr -d '\n'`;
  $tunnel_ip = preg_replace('~[\r\n]+~', '', $tunnel_ip);
  if ($tunnel_ip === false || $tunnel_ip == ''){
    fprintf($log, "%s, no tunnel IP found.\n", $func);
    return;
  }
  fprintf($log, "%s, tunnel IP $tunnel_ip.\n", $func);

  $query = "SELECT system_id FROM bp.systems WHERE role LIKE '%Replication Source'";
  $retval = pg_query($localDbConn, $query);
  if ($retval  && pg_num_rows($retval) > 0) {
     $resultArr = pg_fetch_all($retval);
  }

  $resultArr = pg_fetch_all($retval);
  foreach($resultArr as $result) {
    if ($debug) { $x=json_encode($result);fprintf($log, "%s, results: %s\n", $func, $x); }
    $system_id = $result['system_id'];
    $running = isDlTunnelRunning($system_id);
    if ($running === false) {
      continue;
    }
    // check the ip
    // Sample output: Success|//192.168.123.22/rflr_tunnel_dl_sid_2/
    $prefix="Success|//";
    $suffix="/rflr_tunnel_dl_sid_".$system_id."/";
    $path = `$bpbin/cmc_proxy_mount get_tunnel_path $system_id`;
    $path = preg_replace('~[\r\n]+~', '', $path);
    fprintf($log, "%s, raw tunnel_path for %d is %s \n", $func, $system_id, $path);
    if (strncmp($path, $prefix, strlen($prefix)) !== 0) {
      fprintf($log, "%s, malformed path for source system %d, not using.\n", $func, $system_id, $path);
      continue;
    }
    $path = ltrim($path,$prefix);
    fprintf($log, "%s, ltrim tunnel_path for source system %d is %s \n", $func, $system_id, $path);
    $sys_ip = rtrim($path,$suffix);
    fprintf($log, "%s, IP for source system %d is %s \n", $func, $system_id, $sys_ip);
    if ($sys_ip === '') {
      fprintf($log, "%s, malformed IP for source system %d, not using.\n", $func, $system_id, $sys_ip);
      continue;
    }
    if ($tunnel_ip !== $sys_ip) {
      fprintf($log, "%s, Mismatch!! IP for source system %d, does not match tunnel IP %s, removing tunnel.\n",
        $func, $system_id, $sys_ip, $tunnel_ip);
      $destroyed = bp_destroy_proxy_tunnel($system_id);
      if ($destroyed === true) {
        fprintf($log, "%s, source-side tunnel for system_id %d removed due to mismatch IP.\n", $func, $system_id);
      }
    }
    else {
      fprintf($log, "%s, IP for source system %d, and tunnel IP match.\n", $func, $system_id);
    }
  }
}

function destroyDeadSourceTunnels($localDbConn)
{
  global $bpbin;
  global $log;
  global $debug;

  // UC0: remove any source tunnels that cannot be validated to the target's path

  $func = "destroyDeadSourceTunnels()";
  fprintf($log, "%s, start.\n", $func);

  $tunnel_verify_file =".tunnel_created";

  // checking source tunnels...
  // for each localhost, teardown an active tunnel
  $query = "SELECT system_id FROM bp.systems WHERE is_local_system ='t'";
  fprintf($log, "%s, query: %s\n", $func, $query);
  $retval = pg_query($localDbConn, $query);
  if ($retval  && pg_num_rows($retval) > 0) {
    $resultArr = pg_fetch_all($retval);
    foreach($resultArr as $result) {
      if ($debug) { $x=json_encode($result);fprintf($log, "%s, results: %s\n", $func, $x); }
      $system_id = $result['system_id'];
      $running = isDlTunnelRunning($system_id);
      if ($running === false) {
        continue;
      }

      // validate the path
      // Sample output: Success|/backups/_rflr/download_sid_1
      $prefix="Success|";
      $path = `$bpbin/cmc_proxy_mount get_tunnel_path $system_id`;
      $path = preg_replace('~[\r\n]+~', '', $path);
      fprintf($log, "%s, raw tunnel_path for %d is %s \n", $func, $system_id, $path);
      if (strncmp($path, $prefix, strlen($prefix)) !== 0) {
        fprintf($log, "%s, malformed path for source system %d, not using.\n", $func, $system_id, $path);
        continue;
      }
      $path = ltrim($path,$prefix);
      fprintf($log, "%s, ltrim tunnel_path for source system %d is %s \n", $func, $system_id, $path);

      $verifyfile = $path."/".$tunnel_verify_file;
      if ( file_exists("${verifyfile}") ) {
        fprintf($log,"%s, verified path to target: %s\n", $func, $verifyfile);
        continue;
      }

      $destroyed = bp_destroy_proxy_tunnel($system_id);
      if ($destroyed === true) {
        fprintf($log, "%s, source-side tunnel for system_id %d removed.\n", $func, $system_id);
      }
      else {
        fprintf($log, "%s, source-side tunnel for system_id %d NOT removed.\n", $func, $system_id);
      }
    }
  }
  else {
    fprintf($log, "%s, query: failed, or no source results, %s\n", $func, $retval);
  }
}

function destroyOrphanedSymlinks($localDbConn, $force=false)
{
  global $log;
  global $debug;

  $func = "destroyOrphanedSymlinks()";
  fprintf($log, "%s, start.\n", $func);

  // UC0: only delete orphans that do not have a valid system_id (force=false)
  // UC1: clean all, including orphans that have a valid system_id (force=true)
  // orphaned symlinks can occur when the system is rebooted.

  $tunnel_verify_file =".tunnel_created";
  $downloadsDir       = "/var/www/html/downloads/";

  // 2 Types of Symlinks, this routine checks one at a time:
  // - Tunnel URL DL: /backups/_rflr/download_sid_1/
  // -  Local URL DL: /backups/D2DBackups/_rflr/sid_1/

  // TYPE 1 Check: Tunnel URL DL: /backups/_rflr/download_sid_1/
  fprintf($log,"%s, TYPE 1 Check begin +++++++++++++++++++++++++++++++\n", $func);
  $downloadSidStr = "download_sid_";

  $dirResults = getDirsAndLinks($downloadsDir); // getDirs actually gets files and dirs!!
  foreach ($dirResults as $fullfilepath) {
    if ( !is_link("${fullfilepath}") )  {
      fprintf($log,"%s, skipping non-link file %s\n", $func, $fullfilepath);
      continue;
    }
    $theEndOfTheLink = readlink($fullfilepath);
    fprintf($log,"%s, end of the link is: %s\n", $func, $theEndOfTheLink);
    if (!strstr($theEndOfTheLink, $downloadSidStr)) {
      // not our style, skip it
      fprintf($log,"%s, found a link, not our type, skipping it..%s\n", $func, $fullfilepath);
      continue;
    }
    fprintf($log,"%s, file link name: %s\n", $func, $fullfilepath);

    // To locate the cmc_proxy_mount, the sid id is required, which can be peeled from the link
    // $theEndOfTheLink is something like: "/backups/_rflr/download_sid_1"
    // Strip off any trailing /, then we're left with a sid_#
    $theEndOfTheLink = rtrim($theEndOfTheLink, '/');
    $sid = end(explode($downloadSidStr, $theEndOfTheLink));
    if (!is_numeric($sid) || $sid < 1) {
      fprintf($log,"%s, The extracted sid value was invalid: %s, skipping\n", $func, $sid);
      continue;
    }
    fprintf($log,"%s, The extracted sid value: %d\n", $func, $sid);

    $running = isDlTunnelRunning($sid);
    if ($running === false) {
      fprintf($log,"%s, The tunnel with sid %d was not running, removing orphaned symlink.\n", $func, $sid);
      unlink($fullfilepath);
      continue;
    }
    fprintf($log,"%s, The tunnel with sid %d IS running.\n", $func, $sid);

    // Running dl url, remove it or leave it?

    $verified   = false;
    $verifyfile = $fullfilepath."/".$tunnel_verify_file;
    if ( file_exists("${verifyfile}") ) {
      fprintf($log,"%s, verified file link name: %s\n", $func, $fullfilepath);
      $verified = true;
    }

    // check for a deadlink. If dead, remove any running dl url proxy
    $deadLink = false;
    if (!file_exists($fullfilepath)) {
      fprintf($log,"%s, deadlink found for file link name: %s\n", $func, $fullfilepath);
      $deadLink = true;
    }

    if ($force === true || $verified === false || $deadLink === true) {
      fprintf($log,"%s, removing proxy tunnel: %s, force: %d, verified: %d, deadlink: %d\n",
        $func, $fullfilepath, $force, $verified, $deadlink);
      $destroyed = bp_destroy_proxy_tunnel($sid);
      if ($destroyed === true) {
        fprintf($log, "%s, source-side tunnel for system_id %d removed.\n", $func, $sid);
      }
      else {
        fprintf($log, "%s, source-side tunnel for system_id %d NOT removed.\n", $func, $sid);
      }
      continue;
    }
    fprintf($log,"%s, The tunnel with sid %d is left in a running state, including the symlink.\n", $func, $sid);
  }

  // TYPE 2 Check: Local URL DL: /backups/D2DBackups/_rflr/sid_1/
  fprintf($log,"%s, TYPE 2 Check begin +++++++++++++++++++++++++++++++\n", $func);
  $downloadSidStr = "_rflr/sid_";

  $dirResults = getDirsAndLinks($downloadsDir); // getDirs actually gets files and dirs!!
  foreach ($dirResults as $fullfilepath) {
    if ( !is_link("${fullfilepath}") )  {
      fprintf($log,"%s, skipping non-link file %s\n", $func, $fullfilepath);
      continue;
    }

    $theEndOfTheLink = readlink($fullfilepath);
    fprintf($log,"%s, end of the link is: %s\n", $func, $theEndOfTheLink);
    if (!strstr($theEndOfTheLink, $downloadSidStr)) {
      // not our style, skip it
      fprintf($log,"%s, found a link, not our type, skipping it..%s\n", $func, $fullfilepath);
      continue;
    }
    fprintf($log,"%s, file link name: %s\n", $func, $fullfilepath);

    // To locate the cmc_proxy_mount, the sid id is required, which can be peeled from the link
    // $theEndOfTheLink is something like: "/backups/_rflr/download_sid_1"
    // Strip off any trailing /, then we're left with a sid_#
    $theEndOfTheLink = rtrim($theEndOfTheLink, '/');
    $sid = end(explode($downloadSidStr, $theEndOfTheLink));
    if (!is_numeric($sid) || $sid < 1) {
      fprintf($log,"%s, The extracted sid value was invalid: %s, skipping\n", $func, $sid);
      continue;
    }
    fprintf($log,"%s, The extracted sid value: %d\n", $func, $sid);

    $running = isDlUrlRunning($sid);
    if ($running === false) {
      fprintf($log,"%s, The dl url with sid %d was not running, removing orphaned symlink.\n", $func, $sid);
      unlink($fullfilepath);
      continue;
    }
    fprintf($log,"%s, The dl url with sid %d IS running.\n", $func, $sid);

    // No verify step for local DL Url

    // Running dl url, remove it or leave it?

    // check for a deadlink. If dead, remove any running dl url proxy
    $deadLink = false;
    if (!file_exists($fullfilepath)) {
      fprintf($log,"%s, deadlink found for file link name: %s\n", $func, $fullfilepath);
      $deadLink = true;
    }

    if ($deadLink === true || $force === true) {
      fprintf($log,"%s, removing dl url: %s, deadLink: %d, force: %d\n", $func, $fullfilepath, $deadLink, $force);
      $destroyed = bp_destroy_dl_url($sid);
      if ($destroyed === true) {
        fprintf($log, "%s, source-side dl url for system_id %d removed.\n", $func, $sid);
      }
      else {
        fprintf($log, "%s, source-side dl url for system_id %d NOT removed.\n", $func, $sid);
      }
      continue;
    }
    fprintf($log,"%s, The dl url with sid %d is left in a running state, including the symlink.\n", $func, $sid);
  }
}

function handleDownloadTunnelsAndLinks($localDbConn)
{
  // A download tunnel is needed for making available a directory on the target via a URL on the source.
  // Building the tunnel has proven to be problematic as some timing of the smb refresh on the target
  // seems to mess things up. Also, reboots cause pieces of tunnels (the proxy state stuff for one) to
  // be removed. Therefore, this watchdog determines if tunnels are needed based on replication configuration
  // and builds the target-side of the tunnel within. It does not build source-side tunnels up front, but
  // if one is found, it verifies its connectivity to the target. If verification fails, the source tunnel
  // is deleted.
  // 
  // URL symlinks, are also tracked by this handler to cleanup any expired, orphaned ones. The sequence
  // for cleanup starts when the directory content is removed (expires) from the target, and the source
  // side link becomes dead.
  // 
  // Note, the logic to cleanup orphaned tunnels (replication sources removed) is TBD.

  global $log;
  global $debug;

  $func = "handleDownloadTunnelsAndLinks()";
  fprintf($log, "%s, start.\n", $func);

  // check state of tunnels, in or out
  $createDLTunnels = get_ini_bool_value("SelfService", "CreateDownloadTunnels");

  // target or crossvault?
  $ret = bp_local_system_is_vault();
  if ($ret === true) {
    $type = 'target';
    if ($createDLTunnels === false) {
      destroyOrphanedTunnels($localDbConn, true); // clean all, including orphans that have a valid system_id
      destroyAllTunnels($localDbConn, $type);
    }
    else {
      destroyOrphanedTunnels($localDbConn, false); // only delete orphans that do not have a valid system_id
      checkTargetDlTunnelAdapter($localDbConn); // must run before handleTargetDlTunnels, because it may delete an obsolete tunnel
      handleTargetDlTunnels($localDbConn);
    }
  }

  $type = 'source';
  if ($createDLTunnels === false) {
      destroyAllTunnels($localDbConn, $type);
      destroyOrphanedSymlinks($localDbConn, true); // clean all, including orphans that have a valid system_id
  }
  else {
      destroyDeadSourceTunnels($localDbConn); // tunnels that cannot be validated to the target
      destroyOrphanedSymlinks($localDbConn, false); // only delete orphans links that cannot be verified
  }
}

function doCleanup()
{
  global $log;
  global $debug;
  global $loc_conn_str;

  // already a watchdog running?
  $pid_file = "/var/run/rflr_manage_wd.pid";
  if (file_exists( "$pid_file" )){
    $prevPid = file_get_contents($pid_file);
    $running = posix_kill($prevPid,0);
    if ($running === true) {
      #echo "Already running \n";
      exit;
    }
  }
  // update the pid file with our process
  file_put_contents($pid_file, getmypid());

  setupLogging();

  // The primary intent of this function is to remove all the temporary directories created during downloads
  // per local or remote selective file restores and FLRs. Ideally every task would post progress and perform
  // directory cleanup. However, for the normal selective file download case, the unique temporary directories
  // linger after the job. These are covered in UC1. Jobs can also be aborted at various stages leaving the 
  // job status in the rflr_jobs table in a 'start' state. Jobs should transition from 'start', to 'downloading'.
  // But no 'done' state is recorded. A guess is made as to when to remove the tmp directory (when empty).
  // Then there are the FLR spin ups, which do not post anything to the rflr_jobs table. Again, a guess is made
  // when these directories can be purged.

  $func = "doCleanup()";
  fprintf($log, "%s, start.\n", $func);

  # Get a local db connection, used by the watchdog
  $localDbConn = pg_connect($loc_conn_str);
  if ($localDbConn === false) {
    fprintf($log, "%s, Could not connect to the database.\n", $func);
    return;
  }

  // rflr_jobs status is 'downloading'
  // UC1: and the dir is empty
  // UC2: and the dir is not empty
  handleDownloadingState($localDbConn);

  // rflr_jobs status is 'start'
  // UC3: if restore job, continue
  // UC4: no restore job, if the dir is empty, cleanup
  // UC5: no restore job, the dir is not empty, how long to wait for a download op?
  handleStartState($localDbConn);

  // UC6: lingering tmp directories
  // if empty, candidates for cleanup
  cleanTempDirs($localDbConn);

  // Last chance
  purgeLastChance($localDbConn);

  // Manage the Download Tunnels and URL Symlinks
  // Note: purgeLastChance cleans old tunnel jobs and dirs first.
  // This function ensures tunnels and links are up and operational.
  handleDownloadTunnelsAndLinks($localDbConn);

  pg_close($localDbConn);
}

#####################################################################
# Main program starts here
# Parse command line

$opt = 0;
$shortopts  = "";
$shortopts .= "c:"; // create dir, required starting dir path
$shortopts .= "D";  // Debug option
$shortopts .= "h";  // help option
$shortopts .= "w";  // run watchdog, cleanup option
$shortopts .= "z:"; // create zip, required complete dir path
$shortopts .= "f:"; // create zip files, required complete zip and dir path and filenames
$shortopts .= "g:"; // add up the size in bytes of the provided file list
$shortopts .= "d";  // create zip option, delete files option
$shortopts .= "z";  // zip dir, only for create zip file option
$shortopts .= "s";  // start dir, only for create zip file option
$shortopts .= "n";  // filenames, only for create zip file option
$shortopts .= "e";  // error if filename is a dir, create zip file option

$longopts = array(
"create_dir:",
"Debug",
"help",
"watchdog",
"create_zip:",
"create_zip_files",
"get_file_size",
"delete_files",
"zip_dir:",
"start_dir:",
"filenames:",
"error_if_dir",
);

$options = getopt($shortopts, $longopts);
//var_dump($options);
foreach (array_keys($options) as $opts) switch ($opts) {
  case 'c':
    //$start_dir = (isset($options['c']) ? $options['c'] : '');
    // the short options don't work, require long option
    UsageAndExit();
  break;
  case 'create_dir':
    $createDir = true;
    $start_dir = (isset($options['create_dir']) ? $options['create_dir'] : '');
    $opt++;
  break;
  case 'd':
  case 'delete_files':
    $zipDelFiles = true;
  break;
  case 'D':
  case 'Debug':
    $debug = true;
  break;
  case 'e':
  case 'error_if_dir':
    $errorIfDir = true;
  break;
  case 'h':
  case 'help':
    UsageAndExit();
  break;
  case 'w':
  case 'watchdog':
    $runWatchdog = true;
    $opt++;
  break;
  case 'z':
    // $start_dir = (isset($options['z']) ? $options['z'] : '');
    // the short options don't work, require long option
    UsageAndExit();
  break;
  case 'create_zip':
    $createZip = true;
    $start_dir = (isset($options['create_zip']) ? $options['create_zip'] : '');
    $opt++;
  break;
  case 'create_zip_files':
    $createZipFiles = true;
    $opt++;
  break;
  case 'get_file_size':
    $getFileSize = true;
    $opt++;
  break;
  case 'zip_dir':
    $zipDir = (isset($options['zip_dir']) ? $options['zip_dir'] : '');
  break;
  case 'start_dir':
    $startDir = (isset($options['start_dir']) ? $options['start_dir'] : '');
  break;
  case 'filenames':
    $filename_str = (isset($options['filenames']) ? $options['filenames'] : '');
  break;
  default:
    UsageAndExit();
}

# Must have exactly 1 option specified
if ($opt != 1) {
  UsageAndExit();
}

# create Zip files requires several arugments.
# It should also provide a start_dir but the code can zip things up w/o one. 
if ($createZipFiles == true && ($zipDir == "" || $filename_str == "")){
  UsageAndExit();
}

if ($getFileSize == true && $filename_str == ""){
  UsageAndExit();
}

if ($runWatchdog !== true) {
  setupLogging();
  // Note, the watchdog will first check for another wd instance running, then setupLogging.
}

# need bypass_cookie() for bp calls, e.g. bp_get_local_system_id() and all the tunnel work
bp_bypass_cookie(3);

if ($createDir === true) {
  $echoTrue = true;
  $result = doCreateDir($start_dir, $echoTrue);
}
else if ($createZip === true) {
  $result = doCreateZip($start_dir, $zipDelFiles);
}
else if ($createZipFiles === true) {
  $filenames = explode(",",$filename_str);
  if ($startDir == ""){
    fprintf($log, "WARNING: startDir is null, full FLR restore paths will be leveraged.\n");
  }
  $result = doCreateZipFiles($zipDir, $startDir, $filenames, $errorIfDir);
}
else if ($getFileSize === true) {
  $filenames = explode(",",$filename_str);
  $result = doGetFileSize($filenames);
}
else if ($runWatchdog === true) {
  doCleanup();
}
else {
  UsageAndExit();
}

$msg = "### $program_name completed on ".date('Y-m-d H:i:s');
msg_and_exit($msg, 0);

?>
